# pyright: reportPrivateUsage=false

import threading
import time

import pytest

from escudeiro.data import data
from escudeiro.ds import ThreadPool


@data(frozen=False)
class MockResource:
    state: int
    active: bool

    def close(self):
        time.sleep(0.1)
        self.active = False

    @classmethod
    def create(cls, state: int):
        time.sleep(0.1)
        return cls(state, True)


def get_factory():
    state = 1

    def _factory():
        nonlocal state
        result = MockResource.create(state)
        state += 1
        return result

    return _factory


@pytest.fixture
def thread_pool():
    pool = ThreadPool(get_factory(), MockResource.close)
    yield pool
    pool.dispose()


def test_pool_acquire(thread_pool: ThreadPool[MockResource]):
    resource = thread_pool.acquire()

    assert resource.state == 1


def test_pool_release(thread_pool: ThreadPool[MockResource]):
    resource = thread_pool.acquire()

    thread_pool.release(resource)

    next_resource = thread_pool.acquire()

    assert next_resource is resource


def test_pool_prefill(thread_pool: ThreadPool[MockResource]):
    thread_pool.prefill(3)

    assert thread_pool.resources.qsize() == 3

    thread_pool.prefill(1000)

    assert thread_pool.resources.qsize() == 10


def test_dispose():
    pool = ThreadPool(get_factory(), MockResource.close)

    pool.prefill(3)
    resources = [pool.acquire() for _ in range(pool.resources.qsize())]
    for item in resources:
        pool.release(item)
    pool.dispose()

    for item in resources:
        assert not item.active


def test_thread_pool_racing_condition():
    pool = ThreadPool(get_factory(), MockResource.close, pool_size=2)

    # Acquire the only thread available
    t1 = pool.acquire()

    # Set up a timer to release the thread after a short delay
    timer = threading.Timer(0.5, pool.release, args=[t1])
    timer.start()

    # At the same time, try to acquire a new thread
    # This should result in a race condition
    t2 = pool.acquire()

    # Wait for the timer to release the first thread
    timer.join()

    # Both threads should be different objects
    assert t1 != t2

    # The pool should still only have one available thread
    assert pool._available == 1

    # Release the second thread
    pool.release(t2)

    # The pool should now have two available threads
    assert pool._available == 2


def test_thread_pool_recycle():
    # Create a pool with a pool_recycle timeout of 0.2 seconds
    pool = ThreadPool(
        get_factory(), MockResource.close, pool_size=2, pool_recycle=0.2
    )

    # Acquire a resource and mark its state
    t1 = pool.acquire()
    state = t1.state

    # Release the resource and immediately acquire a new one
    pool.release(t1)
    t2 = pool.acquire()

    # The new resource should have the same state from the previous one
    # if not expired
    assert t2.state == state
    pool.release(t2)

    # Wait for the pool_recycle timeout to expire
    time.sleep(0.3)

    # Acquire a new resource and check its state
    t3 = pool.acquire()

    # The new resource should be recycled and have a new state
    assert t3.state != state
    assert not t1.active

    # Release all resources and dispose of the pool
    pool.release(t3)
    pool.dispose()



from collections.abc import Callable, Sequence
from types import EllipsisType
from typing import Any, ClassVar, Self, overload, override

from .utils.factory import factory, is_factory
from .utils.typedef import MISSING, DisassembledType, TypeNode

type BoolOrCallable = bool | Callable[[Any], Any]


class Field:
    __slots__: ClassVar[Sequence[str]] = (
        "name",
        "annotation",
        "kw_only",
        "default",
        "alias",
        "eq",
        "order",
        "init",
        "hash",
        "repr",
        "asdict_",
        "fromdict",
        "inherited",
        "ref",
    )

    def __init__(
        self,
        name: str,
        annotation: DisassembledType,
        kw_only: bool,
        default: Any,
        alias: str,
        eq: BoolOrCallable,
        order: BoolOrCallable,
        init: bool,
        hash: BoolOrCallable,
        repr: bool | Callable[[Any], str],
        asdict_: Callable[[Any], Any] | None,
        fromdict: Callable[[Any], Any] | None,
        inherited: bool = False,
        ref: Callable[[Any], Any] | None = None,
    ) -> None:
        self.name: str = name
        self.annotation: DisassembledType = annotation
        self.kw_only: bool = kw_only
        self.default: Any = default
        self.alias: str = alias
        self.eq: BoolOrCallable = eq
        self.order: BoolOrCallable = order
        self.init: bool = init
        self.hash: BoolOrCallable = hash
        self.repr: BoolOrCallable = repr
        self.asdict_: Callable[[Any], Any] | None = asdict_
        self.fromdict: Callable[[Any], Any] | None = fromdict
        self.inherited: bool = inherited
        self.ref: Callable[[Any], Any] | None = ref

    @property
    def argname(self):
        return self.alias or self.name

    @property
    def has_alias(self) -> bool:
        return self.alias != self.name

    @property
    def origin(self) -> type | None:
        return self.annotation.origin

    @property
    def args(self) -> Sequence[type]:
        return self.annotation.args

    @property
    def declared_type(self) -> type:
        return self.annotation.type_  # pyright: ignore[reportReturnType]

    @property
    def has_type_vars(self) -> bool:
        return bool(self.annotation.type_vars)

    @property
    def node(self) -> TypeNode:
        return self.annotation.typenode

    @property
    def field_type(self) -> type:
        return self.origin or self.declared_type

    @property
    def has_default(self) -> bool:
        return self.default is not MISSING and not is_factory(self.default)

    @property
    def has_default_factory(self) -> bool:
        return is_factory(self.default)

    @property
    def allow_none(self) -> bool:
        return None in self.args

    @override
    def __repr__(self) -> str:
        default_name = (
            self.default.__name__ if self.has_default_factory else self.default
        )
        return "Field(" + (
            ", ".join(
                (
                    f"name={self.name}",
                    f"annotation={self.annotation}",
                    f"default={default_name}",
                    f"kw_only={self.kw_only}",
                    f"alias={self.alias}",
                    f"eq={self.eq}",
                    f"order={self.order}",
                    f"init={self.init}",
                    f"hash={self.hash}",
                    f"repr={self.repr}",
                    f"asdict_={self.asdict_}",
                    f"fromdict={self.fromdict}",
                    f"transform={self.ref}",
                    f"inherited={self.inherited}",
                )
            )
            + ")"
        )

    def asdict(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def duplicate(self, **overload: Any):
        return type(self)(**self.asdict() | overload)

    def inherit(self) -> Self:
        return self.duplicate(inherited=True)


class FieldInfo:
    __slots__: ClassVar[Sequence[str]] = (
        "default",
        "kw_only",
        "alias",
        "eq",
        "order",
        "init",
        "hash",
        "repr",
        "asdict_",
        "fromdict",
        "ref",
    )

    def __init__(
        self,
        default: Any,
        alias: str,
        kw_only: bool,
        eq: BoolOrCallable,
        order: BoolOrCallable,
        init: bool,
        hash: BoolOrCallable,
        repr: bool | Callable[[Any], str],
        asdict_: Callable[[Any], Any] | None,
        fromdict: Callable[[Any], Any] | None,
        ref: Callable[[Any], Any] | None = None,
    ) -> None:
        self.default: Any = default
        self.kw_only: bool = kw_only
        self.alias: str = alias
        self.eq: BoolOrCallable = eq
        self.order: BoolOrCallable = order
        self.init: bool = init
        self.hash: BoolOrCallable = hash
        self.repr: bool | Callable[[Any], str] = repr
        self.asdict_: Callable[[Any], Any] | None = asdict_
        self.fromdict: Callable[[Any], Any] | None = fromdict
        self.ref: Callable[[Any], Any] | None = ref

    def asdict(self):
        return {key: getattr(self, key) for key in self.__slots__}

    def duplicate(self, **overload: Any):
        return FieldInfo(**self.asdict() | overload)

    def build[F: Field](self, field_cls: type[F], **extras: Any) -> F:
        return field_cls(**self.asdict() | extras)


@overload
def field(
    *,
    default_factory: Callable[[], Any],
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    init: bool = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
) -> Any: ...


@overload
def field(
    *,
    default: Any,
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    init: bool = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
) -> Any: ...


@overload
def field(
    *,
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    init: bool = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
) -> Any: ...


def field(
    *,
    default: Any = ...,
    default_factory: Callable[[], Any] | EllipsisType = ...,
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    init: bool = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
) -> Any:  # sourcery skip: instance-method-first-arg-name
    """
    Declare metadata for a gyver-attrs field.

    :param default: The default value of the field.
    :param default_factory: A callable that returns the default value of the field.
    :param alias: The alternative name for the field.
    :param kw_only: Whether the field should be a keyword-only parameter in
    the generated constructor.
    :param eq: Whether the field should be used in the equality comparison of instances
    of the class.
               If a callable is passed, it will be used to compare the field values.
    :param order: Whether the field should be used in rich comparison ordering
    of instances of the class.
                 If a callable is passed, it will be used to compare the field values.
    """
    if not isinstance(default_factory, EllipsisType):
        default = factory(default_factory)
    return FieldInfo(
        default=default if default is not Ellipsis else MISSING,
        alias=alias,
        kw_only=kw_only,
        eq=eq,
        order=order,
        init=init,
        hash=hash,
        repr=repr,
        asdict_=asdict,
        fromdict=fromdict,
        ref=None,
    )


@overload
def private(
    *,
    initial_factory: Callable[[], Any],
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
): ...


@overload
def private(
    *,
    initial: Any,
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
    ref: Callable[[Any], Any] | None = None,
): ...


@overload
def private(
    *,
    ref: Callable[[Any], Any],
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
): ...


def private(
    *,
    initial: Any = ...,
    initial_factory: Callable[[], Any] | EllipsisType = ...,  # type: ignore
    alias: str = "",
    kw_only: bool = False,
    eq: BoolOrCallable = True,
    order: BoolOrCallable = True,
    hash: BoolOrCallable = True,
    repr: bool | Callable[[Any], str] = True,
    asdict: Callable[[Any], Any] | None = None,
    fromdict: Callable[[Any], Any] | None = None,
    ref: Callable[[Any], Any] | None = None,
) -> FieldInfo:
    """Declare metadata for a field not added to init.
    :param initial: The initial value of the field.
    :param initial_factory: A callable that returns the initial value of the field.
    Parameters are the same as in info."""
    if (
        isinstance(initial, EllipsisType)
        and isinstance(initial_factory, EllipsisType)
        and ref is None
    ):
        raise ValueError("No initial value provided")
    if not isinstance(initial_factory, EllipsisType):
        initial = factory(initial_factory)
    return FieldInfo(
        default=initial if initial is not Ellipsis else MISSING,
        alias=alias,
        kw_only=kw_only,
        eq=eq,
        order=order,
        init=False,
        hash=hash,
        repr=repr,
        asdict_=asdict,
        fromdict=fromdict,
        ref=ref,
    )


default_field_info: FieldInfo = field()



import typing
from collections.abc import Callable
from typing import Self, final

from typing_extensions import TypeIs


@final
class Factory[**P, T]:
    __slots__ = ("func", "_initialized")

    def __new__(cls, func: Callable[P, T]) -> Self:
        if isinstance(func, cls):
            return func

        self = object.__new__(cls)
        return self

    def __init__(self, func: Callable[P, T]) -> None:
        if getattr(self, "_initialized", None):
            return

        self.func = func
        self._initialized = True

    def __name__(self):
        return self.func.__name__

    def __call__(self, *args: P.args, **kwds: P.kwargs) -> T:
        return self.func(*args, **kwds)


def factory[**P, T](func: Callable[P, T]) -> Factory[P, T]:
    return Factory(func)


def is_factory(obj: typing.Any) -> TypeIs[Factory]:
    return isinstance(obj, Factory)



from types import TracebackType
from typing import override

from escudeiro.context.context import AsyncContext, Context
from escudeiro.context.interfaces import (
    AtomicAdapter,
    AtomicAsyncAdapter,
)
from escudeiro.data import data
from escudeiro.lazyfields import dellazy, lazyfield


@data(frozen=False)
class AtomicContext[T](Context[T]):
    adapter: AtomicAdapter[T]

    def __post_init__(self):
        Context.__init__(self, self.adapter)

    @lazyfield
    def client(self):
        return self.adapter.new()

    @override
    def acquire(self) -> T:
        with self._lock:
            if self.adapter.is_closed(self.client):
                dellazy(self, type(self).client)
            self.stack += 1
            self.adapter.begin(self.client)
        return self.client

    @override
    def release(self, commit: bool = True):
        with self._lock:
            if self.stack == 1:
                if self.adapter.in_atomic(self.client):
                    if commit:
                        self.adapter.commit(self.client)
                    else:
                        self.adapter.rollback(self.client)
                self.adapter.release(self.client)
                dellazy(self, type(self).client)
            self.stack -= 1

    @override
    def __exit__(
        self,
        *exc: tuple[
            type[BaseException] | None,
            BaseException | None,
            TracebackType | None,
        ],
    ):
        self.release(not any(exc))


@data(frozen=False)
class AsyncAtomicContext[T](AsyncContext[T]):
    adapter: AtomicAsyncAdapter[T]

    def __post_init__(self):
        AsyncContext.__init__(self, self.adapter)

    @override
    async def acquire(self):
        async with self.lock:
            client = await self.client()
            if self.stack == 0:
                await self.adapter.begin(client)
            self.stack += 1
            return client

    @override
    async def release(self, commit: bool = True):
        async with self.lock:
            client = await self.client()
            if self.stack == 1:
                if await self.adapter.in_atomic(client):
                    if commit:
                        await self.adapter.commit(client)
                    else:
                        await self.adapter.rollback(client)
                await self.adapter.release(client)
                await dellazy(self, type(self).client)
            self.stack -= 1

    @override
    async def __aexit__(
        self,
        *exc: tuple[
            type[BaseException] | None,
            BaseException | None,
            TracebackType | None,
        ],
    ):
        await self.release(not any(exc))



from .atomic_ import AsyncAtomicContext, AtomicContext
from .bound import AsyncBoundContext, BoundContext
from .context import AsyncContext, Context
from .helper import atomic
from .interfaces import Adapter, AsyncAdapter, AtomicAdapter, AtomicAsyncAdapter

__all__ = [
    "Context",
    "AsyncContext",
    "Adapter",
    "AsyncAdapter",
    "AtomicAdapter",
    "AtomicAsyncAdapter",
    "AtomicContext",
    "AsyncAtomicContext",
    "BoundContext",
    "AsyncBoundContext",
    "atomic",
]



# pyright: reportInvalidTypeForm=false

import enum
from collections.abc import Collection, Mapping
from datetime import datetime, tzinfo
from pathlib import Path
from typing import Any, ClassVar, Literal, Self

class url:
    """A collection of URL parsing and manipulation utilities."""

    class Query:
        """Represents and manipulates URL query parameters."""

        def __init__(self, querystr: str) -> None:
            """Initialize with a query string (e.g., 'key1=val1&key2=val2')."""

        def add(self, key: str, value: str) -> None:
            """Add a new key-value pair to the query."""

        def get(self) -> Mapping[str, Collection[str]]:
            """Return all parameters as a mapping of keys to value collections."""

        def encode(self) -> str:
            """Return the URL-encoded query string."""

        def add_map(self, map: Mapping[str, str]) -> None:
            """Add multiple parameters from a mapping of key-value pairs."""

        def set_map(self, map: Mapping[str, str]) -> None:
            """Replace all parameters with those from the given mapping."""

        def remove(self, key: str) -> None:
            """Remove the specified parameter by key."""

        def copy(self) -> Self:
            """Create a deep copy of this Query instance."""

        def omit_empty_equal(self) -> str:
            """Return encoded string omitting equals signs for empty values."""

        def sort(self) -> None:
            """Sort parameters alphabetically by key."""

        def first(self) -> Mapping[str, str]:
            """Return first value for each parameter as a mapping."""

        def compare(self, other: Self) -> bool:
            """Compare if two Query instances have identical parameters."""

        @property
        def params(self) -> Mapping[str, Collection[str]]:
            """Access the underlying parameters mapping directly."""

    class Path:
        """Represents and manipulates URL path components."""

        def __init__(self, pathstr: str) -> None:
            """Initialize with a path string (e.g., '/path/to/resource')."""

        def add(self, pathstr: str) -> None:
            """Append a path segment or segments to the current path."""

        def add_path(self, pathobj: Path) -> None:
            """Append a pathlib.Path object's segments to the current path."""

        def clear(self) -> None:
            """Clear all path segments."""

        def is_dir(self) -> bool:
            """Return True if the path ends with a trailing slash."""

        def encode(self) -> str:
            """Return the properly encoded path string."""

        def normalize(self) -> None:
            """Normalize the path by resolving '.' and '..' segments."""

        def copy(self) -> Self:
            """Create a deep copy of this Path instance."""

        def get(self) -> Collection[str]:
            """Return all path segments as a collection."""

        def compare(self, other: Self) -> bool:
            """Compare if two Path instances have identical segments."""

        @property
        def segments(self) -> Collection[str]:
            """Access the path segments directly."""

    class Fragment:
        """Represents and manipulates URL fragments (after '#')."""

        def __init__(self, fragmentstr: str) -> None:
            """Initialize with a fragment string."""

        def encode(self) -> str:
            """Return the properly encoded fragment string."""

        def set(self, fragmentstr: str) -> None:
            """Replace the current fragment with a new value."""

        def copy(self) -> Self:
            """Create a deep copy of this Fragment instance."""

        def compare(self, other: Self) -> bool:
            """Compare if two Fragment instances are identical."""

    class Netloc:
        """Represents and manipulates network location (host:port with auth)."""

        username: str | None
        password: str | None
        host: str
        port: int | None

        def __init__(self, netloc: str) -> None:
            """Parse a network location string (e.g., 'user:pass@host:port')."""

        def encode(self) -> str:
            """Return the properly encoded network location string."""

        def parse(self, netloc: str) -> None:
            """Parse and replace the current network location values."""

        def set(
            self,
            host: str | None = None,
            port: int | None = None,
            username: str | None = None,
            password: str | None = None,
        ) -> None:
            """Set individual network location components."""

        def merge(self, other: Self) -> Self:
            """Return a new Netloc with values merged from another instance."""

        def merge_left(self, other: Self) -> Self:
            """Return a new Netloc with left-biased merged values."""

        def merge_inplace(self, other: Self) -> None:
            """Merge values from another instance into this one."""

        @staticmethod
        def from_args(
            host: str,
            port: int | None = None,
            username: str | None = None,
            password: str | None = None,
        ) -> url.Netloc:
            """Create a Netloc instance from individual components."""

        def copy(self) -> Self:
            """Create a deep copy of this Netloc instance."""

        def compare(self, other: Self) -> bool:
            """Compare if two Netloc instances are identical."""

    class URL:
        """Comprehensive URL parsing and manipulation class."""

        scheme: str
        netloc: url.Netloc
        path: url.Path
        query: url.Query
        fragment: url.Fragment

        def __init__(self, url: str) -> None:
            """Parse a complete URL string into its components."""

        def encode(self, append_empty_equal: bool = True) -> str:
            """Return the properly encoded URL string."""

        def add(
            self,
            path: str | None = None,
            query: Mapping[str, str] | None = None,
            fragment: str | None = None,
            netloc: str | None = None,
            netloc_obj: url.Netloc | None = None,
            scheme: str | None = None,
        ) -> None:
            """Add components to the existing URL."""

        def set(
            self,
            path: str | None = None,
            query: Mapping[str, str] | None = None,
            fragment: str | None = None,
            netloc: str | None = None,
            netloc_obj: url.Netloc | None = None,
            scheme: str | None = None,
        ) -> None:
            """Replace URL components with new values."""

        @staticmethod
        def from_netloc(
            netloc: url.Netloc | None = None,
            username: str | None = None,
            password: str | None = None,
            host: str | None = None,
            port: int | None = None,
        ) -> url.URL:
            """Create a URL from network location components."""

        @staticmethod
        def from_args(
            path: str | None = None,
            query: Mapping[str, str] | None = None,
            fragment: str | None = None,
            netloc: str | None = None,
            netloc_obj: url.Netloc | None = None,
            scheme: str | None = None,
        ) -> url.URL:
            """Create a URL from individual components."""

        def copy(self) -> Self:
            """Create a deep copy of this URL instance."""

        def compare(self, other: Self) -> bool:
            """Compare if two URL instances are identical."""

class strings:
    """String manipulation utilities."""

    @staticmethod
    def replace_all(value: str, replacements: Mapping[str, str]) -> str:
        """Replace all occurrences of keys with their values in the string."""

    @staticmethod
    def replace_by(
        value: str, replacement: str, to_replace: Collection[str]
    ) -> str:
        """Replace all occurrences of any string in to_replace with replacement."""

    @staticmethod
    def to_snake(value: str) -> str:
        """Convert string to snake_case."""

    @staticmethod
    def to_camel(value: str) -> str:
        """Convert string to camelCase."""

    @staticmethod
    def to_pascal(value: str) -> str:
        """Convert string to PascalCase."""

    @staticmethod
    def to_kebab(value: str, remove_trailing_underscores: bool = True) -> str:
        """Convert string to kebab-case."""

    @staticmethod
    def squote(value: str) -> str:
        """Wrap string in single quotes."""

    @staticmethod
    def dquote(value: str) -> str:
        """Wrap string in double quotes."""

    @staticmethod
    def sentence(value: str) -> str:
        """Capitalize first letter and add period if missing."""

    @staticmethod
    def exclamation(value: str) -> str:
        """Capitalize first letter and add exclamation if missing."""

    @staticmethod
    def question(value: str) -> str:
        """Capitalize first letter and add question mark if missing."""

class squire:
    """Serialization and deserialization utilities."""

    @staticmethod
    def make_mapping(obj: Any, by_alias: bool = False) -> Mapping[str, Any]:
        """Convert an object to a mapping of its attributes."""

    @staticmethod
    def deserialize_mapping(
        mapping: Mapping[str, Any], by_alias: bool = True
    ) -> Mapping[str, Any]:
        """Deserialize a mapping with optional alias resolution."""

    @staticmethod
    def deserialize(value: Any, by_alias: bool = True) -> Any:
        """Generic deserialization with optional alias resolution."""

class cronjob:
    """Cron expression utilities."""

    class CronExpr:
        """Represents and evaluates cron expressions."""

        @property
        def timezone(self) -> tzinfo:
            """Get the timezone used for scheduling."""

        @property
        def next_run(self) -> datetime:
            """Get the next scheduled run time."""

        def __init__(self, expression: str, timezone: tzinfo) -> None:
            """Initialize with a cron expression and timezone."""

        def update(self) -> None:
            """Update next_run to the next scheduled occurrence."""
        def update_after(self, after: datetime) -> None:
            """Update next_run to the next scheduled occurrence after 'after'."""

        def matches(self, value: datetime) -> bool:
            """Check if the given datetime matches the cron schedule."""

class filetree:
    """
    Python interface for file tree manipulation implemented in Rust.

    This module provides utilities for creating and manipulating file and directory
    structures in memory. It includes functions for generating Python filenames
    and classes for representing filesystem nodes and trees.
    """

    @staticmethod
    def python_filename(
        filename: str, private: bool = False, dunder: bool = False
    ) -> str:
        """
        Format a string as a Python filename with appropriate prefixes and extension.

        Args:
            filename: Base name of the file without extension
            private: If True, prepends a single underscore to the filename
            dunder: If True, wraps the filename with double underscores

        Returns:
            Formatted Python filename with .py extension

        Raises:
            ValueError: If both private and dunder are True
        """
        ...

    @staticmethod
    def init_file() -> str:
        """
        Generate a standard Python package __init__.py filename.

        Returns:
            The string "__init__.py"
        """
        ...

    class ErrorCodes(enum.Enum):
        """Mapping of possible error codes."""

        InvalidParam: ClassVar[Literal[1]]
        UnableToAcquireLock: ClassVar[Literal[2]]
        InvalidPath: ClassVar[Literal[3]]
        DuplicateFile: ClassVar[Literal[4]]

    class FsNode:
        """
        Represents a node in a file system tree.

        A node can be either a file (with content) or a directory (containing other nodes).
        Thread-safe operations are ensured through internal locking mechanisms.
        """

        def __init__(self, name: str, content: bytes | None = None) -> None:
            """
            Initialize a new filesystem node.

            Args:
                name: Name of the node
                content: If provided, the node is treated as a file with this content;
                        otherwise, it's treated as a directory
            """
            ...

        @property
        def name(self) -> str:
            """
            Get the name of this node.

            Returns:
                The name of the node

            Raises:
                ValueError: If the internal lock cannot be acquired
            """
            ...

        @property
        def content(self) -> bytes | None:
            """
            Get the content of this node if it's a file.

            Returns:
                The content as bytes if node is a file, None if it's a directory

            Raises:
                ValueError: If the internal lock cannot be acquired
            """
            ...

        @property
        def children(self) -> list[filetree.FsNode]:
            """
            Get all child nodes if this node is a directory.

            Returns:
                List of child nodes

            Raises:
                ValueError: If the internal lock cannot be acquired
            """
            ...

        def is_file(self) -> bool:
            """
            Check if this node represents a file.

            Returns:
                True if the node is a file, False if it's a directory

            Raises:
                ValueError: If the internal lock cannot be acquired
            """
            ...

        def contains_shallow(self, name: str) -> bool:
            """
            Check if this directory contains a child with the given name.
            Only checks immediate children, not descendents.

            Args:
                name: Name of the child node to look for

            Returns:
                True if a child with the given name exists, False otherwise

            Raises:
                ValueError: If the internal lock cannot be acquired
            """
            ...

        def get_shallow(self, name: str) -> filetree.FsNode | None:
            """
            Get a child node by name if it exists.
            Only checks immediate children, not descendents.

            Args:
                name: Name of the child node to retrieve

            Returns:
                The child node if found, None otherwise

            Raises:
                ValueError: If the internal lock cannot be acquired
            """
            ...

        def add_child(self, node: filetree.FsNode) -> None:
            """
            Add a child node to this directory.

            Args:
                node: The node to add as a child

            Raises:
                ValueError: If this node is a file, or if the lock cannot be acquired
            """
            ...

        def write_content(self, content: bytes) -> None:
            """
            Write content to this node.

            Args:
                content: the content to be written in bytes

            Raises:
                ValueError: if this node is a directory, or if the lock cannot be acquired.
            """
            ...

        def append_content(self, content: bytes) -> None:
            """
            Append content to this node.

            Args:
                content: the content to be appended in bytes

            Raises:
                ValueError: if this node is a directory, or if the lock cannot be acquired.
            """
            ...

    class FsTree:
        """
        Represents a file system tree structure.

        Provides methods for creating and navigating directories and files within
        the tree structure.
        """

        def __init__(self, basename: str) -> None:
            """
            Initialize a new file system tree with a root directory.

            Args:
                basename: Name of the root directory
            """
            ...

        @staticmethod
        def from_node(root: filetree.FsNode) -> filetree.FsTree:
            """Initialize a new file system tree with a root directory as the node.

            Args:
                root: Root node of the file tree.
            """

        @property
        def root(self) -> filetree.FsNode:
            """
            Get the root node of this tree.

            Returns:
                The root node
            """
            ...

        def create_dir(self, name: str, *path: str) -> filetree.FsNode:
            """
            Create a new directory at the specified path.

            If the path doesn't exist, intermediate directories will be created.
            If the directory already exists, it will be returned.

            Args:
                name: Name of the directory to create
                *path: Path components leading to the parent directory

            Returns:
                The newly created or existing directory node

            Raises:
                ValueError: If a file with the same name exists at the specified location
            """
            ...

        def create_file(
            self, name: str, content: bytes, *path: str
        ) -> filetree.FsNode:
            """
            Create a new file with the specified content at the given path.

            If the path doesn't exist, intermediate directories will be created.

            Args:
                name: Name of the file to create
                content: Content of the file as bytes
                *path: Path components leading to the parent directory

            Returns:
                The newly created file node

            Raises:
                ValueError: If a node with the same name already exists at the specified location
            """
            ...

        def get_node(self, *path: str) -> filetree.FsNode:
            """
            Get a node at the specified path, creating directories as needed.

            Args:
                *path: Path components to navigate

            Returns:
                The node at the specified path

            Raises:
                ValueError: If a path component is a file but not the last component
            """
            ...



import dataclasses
import sys
from collections.abc import Callable, Generator, Mapping, Sequence
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import (
    Any,
    ForwardRef,
    NoReturn,
    Union,  # pyright:ignore[reportDeprecated]
    dataclass_transform,
    get_args,
    overload,
)

from typing_extensions import TypeIs, evaluate_forward_ref

from escudeiro.data.converters import fromdict, squire
from escudeiro.data.schema import (
    DictSchema,
    HasStr,
    HasToString,
    Items,
    ListSchema,
    Schema,
    str_cast,
)
from escudeiro.data.slots import slot
from escudeiro.misc.typex import is_hashable

from .field_ import Field, FieldInfo, field
from .methods import ArgumentType, MethodBuilder, MethodType
from .resolver import FieldsBuilder
from .utils.functions import disassemble_type, indent, remove_left_underscores
from .utils.functions import frozen as freeze
from .utils.typedef import (
    MISSING,
    UNINITIALIZED,
    Descriptor,
    InitOptions,
)

FieldMap = dict[str, Field]


def __dataclass_transform__[T](
    *,
    eq_default: bool,
    order_default: bool,
    kw_only_default: bool,
    frozen_default: bool,
    field_descriptors: tuple[type | Callable[..., Any], ...],
) -> Callable[[T], T]:
    return lambda a: a


@overload
def data[T](
    maybe_cls: None = None,
    /,
    *,
    frozen: bool = True,
    init: bool = True,
    kw_only: bool = False,
    slots: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = True,
    hash: bool | None = None,
    pydantic: bool = False,
    dataclass_fields: bool = False,
    field_class: type[Field] = Field,
    alias_generator: Callable[[str], str] = str,
) -> Callable[[type[T]], type[T]]: ...


@overload
def data[T](
    maybe_cls: type[T],
    /,
    *,
    frozen: bool = True,
    init: bool = True,
    kw_only: bool = False,
    slots: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = True,
    hash: bool | None = None,
    pydantic: bool = False,
    dataclass_fields: bool = False,
    field_class: type[Field] = Field,
    alias_generator: Callable[[str], str] = str,
) -> type[T]: ...


@dataclass_transform(
    order_default=True,
    frozen_default=True,
    kw_only_default=False,
    field_specifiers=(FieldInfo, field),
)
@__dataclass_transform__(  # Support for both formats of dataclass transform
    eq_default=True,
    order_default=True,
    frozen_default=True,
    kw_only_default=False,
    field_descriptors=(FieldInfo, field),
)
def data[T](
    maybe_cls: type[T] | None = None,
    /,
    *,
    frozen: bool = True,
    init: bool = True,
    kw_only: bool = False,
    slots: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = True,
    hash: bool | None = None,
    pydantic: bool = False,
    dataclass_fields: bool = False,
    field_class: type[Field] = Field,
    alias_generator: Callable[[str], str] = str,
) -> Callable[[type[T]], type[T]] | type[T]:
    """
    Decorator function that adds functionality to a data class.

    :param maybe_cls: Optional[type[T]], a type argument that needs to be
    wrapped in the FieldsBuilder.
    :param frozen: bool, whether to create an immutable class or not.
    :param kw_only: bool, whether all params should be kw_only or not.
    :param slots: bool, whether to generate a class using __slots__ or not.
    :param repr: bool, whether to generate a __repr__ method or not.
    :param eq: bool, whether to generate an __eq__ method or not.
    :param order: bool, whether to generate rich comparison methods or not.
    :param hash: bool, whether to generate a __hash__ method or not.
    :param pydantic: bool, whether to generate a __get_validator__ method or
    not to facilitate integration with pydantic.
    :param dataclass_fields: bool, whether to add __dataclass_fields__ with the
    dataclass format. This way, the class becomes a drop-in replacement for dataclasses.
    :param alias_generator: Callable[[str], str], automatic alias using the callable passed
    as parameter.
    **Warning**: dataclass_fields with pydantic=False will fail when trying to use with
    pydantic.

    :return: A callable object that wraps the maybe_cls type argument in a
    class that implements the specified features.
    :rtype: typing.Union[Callable[[type[T]], type[T]], type[T]]
    """

    def wrap(cls: type[T]) -> type[T]:
        fields = (
            FieldsBuilder(
                cls, kw_only, field_class, dataclass_fields, alias_generator
            )
            .from_annotations()
            .build()
        )
        field_map = {field.name: field for field in fields}
        return build_cls(
            cls,
            field_map,
            frozen=frozen,
            kw_only=kw_only,
            slots=slots,
            repr=repr,
            eq=eq,
            order=order,
            hash=hash,
            pydantic=pydantic,
            dataclass_fields=dataclass_fields,
            init=init,
        )

    return wrap(maybe_cls) if maybe_cls is not None else wrap


def build_cls(cls: type, field_map: FieldMap, **opts: Any) -> Any:
    hash = opts["hash"]
    frozen = opts["frozen"]
    slots = opts["slots"]
    clsdict = (
        _get_clsdict(cls, field_map)
        | _get_cls_metadata(cls, opts)
        | _get_init(
            cls,
            field_map,
            {
                "frozen": frozen,
                "slots": slots,
                "init": opts["init"],
            },
        )
        | get_parse_dict(cls, field_map)
        | get_squire_serialize(cls, field_map)
    )

    if slots:
        clsdict |= _get_slots_metadata(cls, field_map)
    if opts["repr"]:
        clsdict |= _get_repr(cls, field_map)
    if opts["eq"]:
        clsdict |= _get_eq(cls, field_map)
        clsdict |= _get_ne(cls)
    if opts["order"]:
        clsdict |= _get_order(cls, field_map)
    if hash or (hash is None and frozen):
        clsdict |= _get_hash(cls, field_map, bool(hash))
    if opts["pydantic"]:
        clsdict |= _get_pydantic_handlers(cls, field_map)
    if opts["dataclass_fields"]:
        clsdict |= _make_dataclass_fields(field_map)
    maybe_freeze = freeze if frozen else lambda a: a
    return maybe_freeze(
        type(cls)(
            cls.__name__,
            cls.__bases__,
            clsdict,
        )
    )


def _get_clsdict(cls: type, field_map: FieldMap):
    return {
        key: value
        for key, value in cls.__dict__.items()
        if key not in (tuple(field_map) + ("__dict__", "__weakref__"))
    } | {"__squire_attrs__": field_map}


def _get_slots_metadata(
    cls: type,
    field_map: FieldMap,
) -> Mapping[str, Any]:
    inherited_slots: dict[str, Any] = {}
    for base_cls in cls.mro()[1:-1]:
        inherited_slots |= {
            name: getattr(base_cls, name)
            for name in getattr(base_cls, "__slots__", ())
        }
    reused_slots = {
        slot: descriptor
        for slot, descriptor in inherited_slots.items()
        if slot in field_map
    }
    slot_names = tuple(
        field for field in field_map if field not in reused_slots
    )
    for value in cls.__dict__.values():
        if (
            _is_descriptor_type(value)
            and value.private_name not in slot_names
            and value.private_name not in inherited_slots
            and value.private_name not in reused_slots
        ):
            slot_names += (value.private_name,)
        if (slots := slot.get_slots(value)) is not None:
            for extra in slots:
                if extra not in slot_names + tuple(reused_slots) + tuple(
                    inherited_slots
                ):
                    slot_names += (extra,)

    return inherited_slots | reused_slots | {"__slots__": tuple(slot_names)}


def _is_descriptor_type(
    obj: Any,
) -> TypeIs[Descriptor]:
    return hasattr(obj, "private_name") and hasattr(obj, "__get__")


def _get_cls_metadata(cls: type, opts: dict[str, Any]):
    return {
        "__qualname__": cls.__qualname__,
        "__source__": cls,
        "__build_opts__": opts,
    }


def _make_setattr(frozen: bool):
    def _setattr(field: str, arg: Any):
        return (
            f"_setattr(self, '{field}', {arg})"
            if frozen
            else f"self.{field} = {arg}"
        )

    return _setattr


def _get_init(cls: type, field_map: FieldMap, opts: InitOptions):
    method_name = "__init__"
    if not opts["init"]:
        method_name = MethodBuilder.make_squire_name(method_name)
    builder = MethodBuilder(
        method_name,
        {
            "attr_dict": field_map,
            "MISSING": MISSING,
            "_setattr": object.__setattr__,
            "UNINITIALIZED": UNINITIALIZED,
        },
    )
    _setattr = _make_setattr(opts["frozen"])
    if hasattr(cls, "__pre_init__"):
        _ = builder.add_scriptline("self.__pre_init__()")
    if not opts["slots"]:
        _ = builder.add_scriptline("_inst_dict = self.__dict__")
    for f in field_map.values():
        field_name = f.name
        arg_name = remove_left_underscores(f.alias)
        if not f.init:
            if f.has_default:
                _ = builder.add_scriptline(
                    _setattr(field_name, f"attr_dict['{field_name}'].default")
                )
            elif f.has_default_factory:
                factory_name = f"__attr_factory_{field_name}"
                _ = builder.add_scriptline(
                    _setattr(field_name, f"{factory_name}()")
                ).add_glob(factory_name, f.default)
            elif f.ref:
                _ = builder.add_scriptline(
                    _setattr(
                        field_name,
                        f"attr_dict['{field_name}'].ref(self)",
                    )
                )
            else:
                _ = builder.add_scriptline(
                    _setattr(field_name, "UNINITIALIZED")
                )
            continue
        if f.has_default:
            arg = f"{arg_name}=attr_dict['{field_name}'].default"

            _ = builder.add_scriptline(_setattr(field_name, arg_name))
        elif f.has_default_factory:
            arg = f"{arg_name}=MISSING"

            init_factory_name = f"__attr_factory_{field_name}"
            for line in (
                f"if {arg_name} is not MISSING:",
                f"    {_setattr(field_name, arg_name)}",
                "else:",
                f"    {_setattr(field_name, f'{init_factory_name}()')}",
            ):
                _ = builder.add_scriptline(line)
            _ = builder.add_glob(init_factory_name, f.default)
        elif f.ref:
            arg = f"{arg_name}=MISSING"
            init_transform_name = f"__attr_transform_{field_name}"
            for line in (
                f"if {arg_name} is not MISSING:",
                f"    {_setattr(field_name, arg_name)}",
                "else:",
                f"    {_setattr(field_name, f'{init_transform_name}(self)')}",
            ):
                _ = builder.add_scriptline(line)
            _ = builder.add_glob(init_transform_name, f.ref)
        else:
            _ = builder.add_scriptline(_setattr(field_name, arg_name))
            arg = arg_name
        _ = builder.add_arg(
            arg,
            ArgumentType.KEYWORD if f.kw_only else ArgumentType.POSITIONAL,
        )
        _ = builder.add_annotation(arg_name, f.declared_type)
    if hasattr(cls, "__post_init__"):
        _ = builder.add_scriptline("self.__post_init__()")
    return builder.build(cls)


def _get_repr(cls: type, field_map: FieldMap):
    fields = []
    globs = {}
    for f in field_map.values():
        if f.repr is False:
            continue
        elif f.repr is True:
            fields.append(f"{f.name}={{self.{f.name}!r}}")
        else:
            field_repr_call = f"__repr_{f.name}"
            globs[field_repr_call] = f.repr
            fields.append(f"{f.name}={{{field_repr_call}(self.{f.name})!r}}")
    fieldstr = ", ".join(fields)
    returnline = f"return f'{cls.__name__}({fieldstr})'"
    return (
        MethodBuilder("__repr__", globs)
        .add_annotation("return", str)
        .add_scriptline(returnline)
        .build(cls)
    )


_othername = "other"


def _get_eq(cls: type, field_map: FieldMap):
    fields_to_compare = {
        name: f for name, f in field_map.items() if f.eq is not False
    }
    builder = MethodBuilder("__eq__").add_arg(
        _othername, ArgumentType.POSITIONAL
    )
    if fields_to_compare:
        return _build_field_comparison(builder, fields_to_compare, cls)
    returnline = "return _object_eq(self, other)"
    return (
        builder.add_glob("_object_eq", object.__eq__)
        .add_annotation("return", bool)
        .add_scriptline(returnline)
        .build(cls)
    )


def _build_field_comparison(
    builder: MethodBuilder, fields_to_compare: FieldMap, cls: type
):
    _ = builder.add_scriptline("if type(other) is type(self):")
    args = []
    for f in fields_to_compare.values():
        arg = f"{{target}}.{f.name}"
        if f.eq is not True:
            glob_name = f"_parser_{f.name}"
            arg = f"{glob_name}({arg})"
            _ = builder.add_glob(glob_name, f.eq)
        args.append(arg)

    fieldstr = "(" + ", ".join(args) + ",)"
    return (
        builder.add_scriptlines(
            indent(
                f"return {fieldstr.format(target='self')} "
                + f"== {fieldstr.format(target=_othername)}",
            ),
            "else:",
            indent("return NotImplemented"),
        )
        .add_annotation("return", bool)
        .build(cls)
    )


def make_unresolved_ref(cls: type, field: Field):
    def _unresolved_ref(_: Any) -> NoReturn:
        raise TypeError(
            "Trying to use class with unresolved ForwardRef for"
            + f" {cls.__qualname__}.{field.name}",
        )

    return _unresolved_ref


def get_parse_dict(cls: type, field_map: FieldMap):
    namespace = {}
    args = []
    alias_args = []
    builder = (
        MethodBuilder(
            "__parse_dict__",
            {
                "deserialize": squire.deserialize,
                "deserialize_mapping": squire.deserialize_mapping,
            },
        )
        .add_arg("alias", ArgumentType.POSITIONAL)
        .add_annotation("alias", bool)
        .add_annotation("return", Mapping[str, Any])
    )
    mod_globalns = sys.modules[cls.__module__].__dict__
    for name, f in field_map.items():
        field_type = f.origin or f.declared_type
        result, resolved = _resolve_forward_ref(
            field_type, cls, f, mod_globalns
        )
        if not resolved:
            _ = builder.add_glob(f"_asdict_{f.name}", result)
            arg = f"'{{name}}': _asdict_{f.name}(self.{name})"
        else:
            arg = _create_argument_for_field(f, field_type, builder.add_glob)
        args.append(arg.format(name=name))
        alias_args.append(arg.format(name=f.alias))
    namespace |= builder.add_scriptline(
        "\n    ".join(
            (
                "if alias:",
                f"    return {{{', '.join(alias_args)}}}",
                f"return {{{', '.join(args)}}}",
            )
        )
    ).build(cls)

    return namespace | MethodBuilder(
        "__iter__", {"todict": squire.deserialize}
    ).add_scriptline("yield from todict(self).items()").build(cls)


def _resolve_forward_ref(
    field_type: Any,
    cls: type,
    field: Field,
    mod_globalns: dict[str, Any],
) -> tuple[Any, bool]:
    if not isinstance(field_type, ForwardRef) or field.asdict_:
        return field_type, True
    try:
        parsed = evaluate_forward_ref(
            field_type,
            globals=mod_globalns,
            locals={cls.__name__: cls},
            type_params=None,
        )
    except NameError:
        return make_unresolved_ref(cls, field), False
    return (
        (disassemble_type(parsed), True)
        if parsed is not None
        else (
            make_unresolved_ref(cls, field),
            False,
        )
    )


def _create_argument_for_field(
    f: Field,
    field_type: type | ForwardRef,
    add_glob: Callable[[str, Callable], Any],
):
    if f.asdict_:
        add_glob(f"_asdict_{f.name}", f.asdict_)
        return f"'{{name}}': _asdict_{f.name}(self.{f.name})"
    elif hasattr(field_type, "__parse_dict__"):
        return f"'{{name}}': self.{f.name}.__parse_dict__(alias)"
    elif not isinstance(field_type, type):
        return f"'{{name}}': self.{f.name}"
    elif issubclass(field_type, list | tuple | set | dict):
        add_glob(f"field_type_{f.name}", field_type)
        return _get_parse_dict_sequence_arg(f)
    else:
        return f"'{{name}}': self.{f.name}"


def _get_parse_dict_sequence_arg(f: Field) -> str:
    field_type = f.origin or f.declared_type
    if not f.args:
        return f"'{{name}}': self.{f.name}"
    elif (
        len(f.args) > 1
        and issubclass(field_type, tuple)
        and (len(f.args) != 2 or f.args[1] is not Ellipsis)  # pyright: ignore[reportUnnecessaryComparison]
    ):
        idx_to_parse = [
            idx
            for idx, item in enumerate(f.args)
            if hasattr(item, "__parse_dict__")
        ]
        if not idx_to_parse:
            return f"'{{name}}': self.{f.name}"
        tuple_args = ", ".join(
            f"self.{f.name}[{idx}]"
            if idx not in idx_to_parse
            else f"self.{f.name}[{idx}].__parse_dict__(alias)"
            for idx, _ in enumerate(f.args)
        )
        return f"'{{name}}': ({tuple_args})"
    elif len(f.args) == 1 or issubclass(field_type, tuple):
        (element_type, *_) = f.args
        if hasattr(element_type, "__parse_dict__"):
            return (
                f"'{{name}}': field_type_{f.name}(x.__parse_dict__(alias)"
                f" for x in self.{f.name})"
            )
        return f"'{{name}}': deserialize(self.{f.name}, alias)"
    elif issubclass(field_type, Mapping):
        return f"'{{name}}': deserialize_mapping(self.{f.name}, alias)"
    else:
        return f"'{{name}}': deserialize(self.{f.name}, alias)"


def _dict_get(
    mapping: Mapping,
    name: str,
    alias: str,
    sentinel: Any,
    default: Any,
    dict_get: Callable[[Mapping, str, Any], Any],
):
    val = dict_get(mapping, alias, sentinel)
    if val is not sentinel:
        return val
    val = dict_get(mapping, name, sentinel)
    if val is not sentinel:
        return val
    if default is not sentinel:
        return default
    raise KeyError(
        f"Key '{name}' or alias '{alias}' not found in mapping: {mapping!r}"
        + " and no default value provided.",
    )


def get_squire_serialize(cls: type, field_map: FieldMap):
    args = []
    builder = (
        MethodBuilder(
            "__squire_serialize__",
            {
                "dict_get": _dict_get,
                "sentinel": object(),
                "_dict_get": dict.get,
            },
        )
        .add_arg("mapping", ArgumentType.POSITIONAL)
        .add_annotation("return", cls)
        .add_annotation("mapping", Mapping[str, Any])
        .set_type(MethodType.CLASS)
    )
    for f in field_map.values():
        field_type = f.origin or f.declared_type
        _ = builder.add_glob(f"_field_type_{f.name}", field_type)
        default_line = "sentinel"
        if f.has_default:
            _ = builder.add_glob(f"_field_type_{f.name}_default", f.default)
            default_line = f"_field_type_{f.name}_default"
        elif f.has_default_factory:
            _ = builder.add_glob(
                f"_field_type_{f.name}_default_factory", f.default
            )
            default_line = f"_field_type_{f.name}_default_factory()"
        get_line = f"dict_get(mapping, {f.name!r}, {f.alias!r}, sentinel, {default_line}, _dict_get)"
        if f.fromdict:
            _ = builder.add_glob(f"_field_type_{f.name}", f.fromdict)
            arg = f"_field_type_{f.name}({get_line})"
        elif hasattr(field_type, "__squire_serialize__"):
            arg = f"_field_type_{f.name}.__squire_serialize__({get_line})"
        elif field_type in (date, datetime):
            arg = f"_field_type_{f.name}.fromisoformat({get_line})"
        elif not isinstance(field_type, type):  # pyright: ignore[reportUnnecessaryIsInstance]
            arg = f"({get_line})"
        elif issubclass(field_type, list | tuple | set | dict):
            arg, globs = _get_gserialize_sequence_arg(f, get_line)
            _ = builder.merge_globs(globs)
        else:
            arg = f"_field_type_{f.name}({get_line})"
        args.append(f"{f.alias}={arg}")
    _ = builder.add_scriptline(f"return cls({', '.join(args)})")
    return builder.build(cls)


def _get_gserialize_sequence_arg(
    field: Field,
    get_line: str,
) -> tuple[str, Mapping[str, Any]]:
    field_type = field.origin or field.declared_type
    globs = {}
    returnline = get_line
    if not field.args:
        pass
    elif (
        len(field.args) > 1
        and issubclass(field_type, tuple)
        and (len(field.args) != 2 or field.args[1] is not Ellipsis)  # pyright: ignore[reportUnnecessaryComparison]
    ):
        if idx_to_parse := [
            idx
            for idx, item in enumerate(field.args)
            if hasattr(item, "__squire_serialize__")
        ]:
            for idx in idx_to_parse:
                globs[f"_elem_type_{field.name}_{idx}"] = field.args[idx]
            tuple_args = ", ".join(
                f"{get_line}[{idx}]"
                if idx not in idx_to_parse
                else f"_elem_type_{field.name}_{idx}."
                + f"__squire_serialize__({get_line}[{idx}])"
                for idx, _ in enumerate(field.args)
            )
            returnline = f"({tuple_args})"
    elif len(field.args) == 1 or issubclass(field_type, tuple):
        (element_type, *_) = field.args
        if hasattr(element_type, "__squire_serialize__"):
            globs[f"_elem_type_{field.name}"] = element_type
            returnline = (
                f"_field_type_{field.name}("
                + f"_elem_type_{field.name}.__squire_serialize__(x)"
                + f" for x in {get_line})"
            )
        else:
            returnline = f"_field_type_{field.name}({get_line})"
    return returnline, globs


def _get_ne(cls: type):
    return (
        MethodBuilder("__ne__")
        .add_arg(_othername, ArgumentType.POSITIONAL)
        .add_annotation("return", bool)
        .add_scriptlines(
            "result = self.__eq__(other)",
            "if result is NotImplemented:",
            indent("return NotImplemented"),
            "else:",
            indent("return not result"),
        )
        .build(cls)
    )


def _get_order(cls: type, field_map: FieldMap):
    payload: dict[str, Any] = {}

    for name, signal in [
        ("__lt__", "<"),
        ("__le__", "<="),
        ("__gt__", ">"),
        ("__ge__", ">="),
    ]:
        payload |= _make_comparator_builder(name, signal, field_map).build(cls)
    return payload


def _get_order_attr_tuple(fields: list[Field]) -> str:
    args = []
    for f in fields:
        arg = f"{{target}}.{f.name}"
        if f.order is not True:
            arg = f"_parser_{f.name}({arg})"
        args.append(arg)

    return f"({', '.join(args)},)"


def _make_comparator_builder(name: str, signal: str, field_map: FieldMap):
    fields = [f for f in field_map.values() if f.order is not False]

    if not fields:
        return (
            MethodBuilder(name, {f"_object_{name}": getattr(object, name)})
            .add_arg(_othername, ArgumentType.POSITIONAL)
            .add_annotation("return", bool)
            .add_scriptline(f"return _object_{name}(self, other)")
        )
    builder = MethodBuilder(
        name,
        {f"_parser_{f.name}": f.order for f in field_map.values()},
    )
    attr_tuple = _get_order_attr_tuple(fields)
    return (
        builder.add_arg(_othername, ArgumentType.POSITIONAL)
        .add_annotation("return", bool)
        .add_scriptlines(
            "if type(other) is type(self):",
            indent(
                "return "
                + f" {signal} ".join(
                    (
                        attr_tuple.format(target="self"),
                        attr_tuple.format(target="other"),
                    )
                ),
            ),
            "return NotImplemented",
        )
    )


def _get_hash(cls: type, fields_map: FieldMap, wants_hash: bool):
    builder = MethodBuilder("__hash__")
    args = ["type(self)"]
    for f in fields_map.values():
        if not f.hash:
            continue
        arg = f"self.{f.name}"
        if f.hash is not True:
            glob = f"_hash_{f.name}"
            arg = f"{glob}({arg})"
            _ = builder.add_glob(glob, f.hash)
        elif not isinstance(f.eq, bool):
            glob = f"_hash_{f.name}"
            arg = f"{glob}({arg})"
            _ = builder.add_glob(glob, f.eq)
        elif not is_hashable(f.declared_type):
            if not wants_hash:
                continue
            raise TypeError(
                f"field type is not hashable: {f.declared_type!r} (field '{f.name}' in {cls.__name__})"
            )
        args.append(arg)

    # if it only contains the class and no field qualifies for hashing
    if len(args) == 1:
        if not wants_hash:
            return {}
        raise TypeError("No hashable field found for class")

    return (
        builder.add_scriptline(f"return hash(({', '.join(args)}))")
        .add_annotation("return", int)
        .build(cls)
    )


def _get_pydantic_handlers(cls: type, fields_map: FieldMap):
    namespace = {}

    # Create Validation Function
    builder = MethodBuilder(
        "__pydantic_validate__", {"fromdict": fromdict}
    ).set_type(MethodType.CLASS)
    namespace |= (
        builder.add_arg("value", ArgumentType.POSITIONAL)
        .add_annotation("value", Any)
        .add_annotation("return", cls)
        .add_scriptlines(
            "if isinstance(value, cls):",
            indent("return value"),
            "try:",
            indent("return fromdict(cls, dict(value))"),
            "except (TypeError, ValueError) as e:",
            indent(
                f"raise TypeError(f'{cls.__name__} expected dict not"
                + " {type(value).__name__}')",
            ),
        )
        .build(cls)
    )

    # Write Get Validators
    namespace |= (
        MethodBuilder("__get_validators__")
        .set_type(MethodType.CLASS)
        .add_scriptline("yield cls.__pydantic_validate__ ")
        .add_annotation("return", Generator[Any, Any, Callable])
        .build(cls)
    )

    # Make modify schema
    builder = (
        MethodBuilder("__modify_schema__")
        .set_type(MethodType.CLASS)
        .add_arg("field_schema", ArgumentType.POSITIONAL)
    )

    cls_schema = Schema(
        cls.__name__,
        "object",
        [f.argname for f in fields_map.values() if f.default is MISSING],
        properties=_generate_schema_lines(fields_map),
    )
    return namespace | builder.add_scriptline(
        f"field_schema.update({cls_schema.to_string()})"
    ).build(cls)


def _generate_schema_lines(fields_map: FieldMap) -> dict[str, HasStr]:
    schemas: dict[str, HasStr] = {}
    for f in fields_map.values():
        field_type = f.field_type
        if current_map := getattr(field_type, "__squire_attrs__", None):
            required = [
                f.argname for f in current_map.values() if f.default is MISSING
            ]
            schemas[f.argname] = Schema(
                field_type.__name__,
                "object",
                required,
                properties=_generate_schema_lines(current_map),
            )
        else:
            schemas[f.argname] = _resolve_schematype(f.field_type, f.args)
    return schemas


def _resolve_schematype(field_type: type, args: Sequence[type]) -> HasToString:
    _type_map = {
        type(None): "null",
        bool: "boolean",
        str: "string",
        float: "number",
        int: "integer",
    }
    if val := _type_map.get(field_type):
        return DictSchema(val)
    if field_type in (list, set, tuple):
        extras = {}
        if args:
            extras["items"] = Items(
                *(_resolve_schematype(arg, get_args(arg)) for arg in args)
            )
        return DictSchema("array", **extras)
    if field_type is dict:
        extras = {}
        if args:
            keyt, valt = args
            if keyt is not str:
                raise TypeError(
                    f"Cannot generate schema for dict with key type {keyt}",
                    keyt,
                )
            extras["additional_properties"] = _resolve_schematype(
                valt, get_args(valt)
            )
        return DictSchema("object", **extras)
    if current_map := getattr(field_type, "__squire_attrs__", None):
        required = [
            f.argname for f in current_map.values() if f.default is MISSING
        ]
        return Schema(
            field_type.__name__,
            "object",
            required,
            properties=_generate_schema_lines(current_map),
        )
    if field_type is Union:  # pyright: ignore[reportDeprecated]
        choices = [_resolve_schematype(arg, get_args(arg)) for arg in args]
        return ListSchema("anyOf", *choices)
    if issubclass(field_type, Enum):
        # remove from parents cls, enum.Enum and object
        # uses only the first of the mro
        parent, *_ = field_type.mro()[1:-2]
        if parent:
            if ft := _type_map.get(parent):
                type_name = ft
            else:
                raise NotImplementedError
        else:
            type_name = "string"
        return Schema(
            field_type.__name__,
            type_name,
            enum=Items(*[f'"{item.value}"' for item in field_type]),
        )
    if issubclass(field_type, datetime):
        return DictSchema("string", format=str_cast("date-time"))
    if issubclass(field_type, date):
        return DictSchema("string", format=str_cast("date"))
    if issubclass(field_type, time):
        return DictSchema("string", format=str_cast("time"))
    if issubclass(field_type, timedelta):
        return DictSchema("number", format=str_cast("time-delta"))
    raise NotImplementedError


def _make_dataclass_fields(fields_map: FieldMap):
    dc_fields = {}
    for f in fields_map.values():
        kwargs: dict[str, Any] = {
            "compare": f.eq or f.order,
            "hash": bool(f.hash),
            "repr": bool(f.repr),
            "init": f.init,
        }
        if f.has_default:
            kwargs["default"] = f.default
        if f.has_default_factory:
            kwargs["default_factory"] = f.default

        dc_field: dataclasses.Field = dataclasses.field(**kwargs)
        dc_field.name = f.name
        dc_field.type = f.declared_type
        dc_field._field_type = dataclasses._FIELD  # pyright: ignore[reportAttributeAccessIssue]
        dc_fields[f.name] = dc_field
    return {"__dataclass_fields__": dc_fields}



import asyncio
import threading

from escudeiro.context.atomic_ import AsyncAtomicContext, AtomicContext
from escudeiro.context.context import AsyncContext, Context
from escudeiro.context.helper import atomic

from .mocks import MockAdapter, MockAsyncAdapter


def test_atomic_context_acquire_release():
    # Create an instance of AtomicContext
    context = AtomicContext(MockAdapter())

    # Test that `acquire` and `release` work as expected
    client = context.acquire()
    assert client.count == 1
    context.release()
    assert client.count == 0


def test_atomic_context_multi_threading():
    adapter = MockAdapter()
    context = atomic(Context(adapter), bound=False)

    def worker():
        with context.open():
            with context.begin() as client:
                assert not client.closed
                client2 = context.acquire()
                assert client == client2
                context.release()

    threads = []
    for _ in range(5):
        t = threading.Thread(target=worker)
        t.start()
        threads.append(t)
    for t in threads:
        t.join()

    assert not context.is_active()
    assert context.stack == 0


def test_bound_context_client_yields_the_same_client_instantiated_in_context():
    adapter = MockAdapter()
    context = Context(adapter)
    bound_context = atomic(context)
    with context.begin() as ctx_client:
        with bound_context.begin() as bnd_ctx_client:
            assert ctx_client is bnd_ctx_client


async def test_atomic_async_context_acquire_release():
    # Create an instance of AtomicAsyncContext
    context = AsyncAtomicContext(MockAsyncAdapter())

    # Test that `acquire` and `release` work as expected
    client = await context.acquire()
    assert client.count == 1
    await context.release()
    assert client.count == 0


def test_atomic_function_works_as_expected_with_context():
    context = Context(MockAdapter())
    with atomic(context) as client:
        assert not client.closed
        assert client.count == 1
    assert client.closed
    assert client.count == 0


async def test_atomic_function_works_as_expected_with_async_context():
    context = AsyncContext(MockAsyncAdapter())
    async with atomic(context) as client:
        assert not client.closed
        assert client.count == 1
    assert client.closed
    assert client.count == 0


async def test_asynccontext_multitask():
    adapter = MockAsyncAdapter()
    context = atomic(AsyncContext(adapter), bound=False)

    async def worker():
        async with context.open():
            async with context.begin() as client:
                assert context.stack == 2
                assert not client.closed
                client2 = await context.acquire()
                assert client == client2
                await context.release()
        assert context.stack == 0
        assert client.closed

    tasks = []
    for _ in range(5):
        tasks.append(asyncio.create_task(worker()))
    _ = await asyncio.gather(*tasks)


async def test_bound_async_context_client_yields_the_same_client_in_context():
    adapter = MockAsyncAdapter()
    context = AsyncContext(adapter)
    bound_context = atomic(context)
    async with context.begin() as ctx_client:
        async with bound_context.begin() as bnd_ctx_client:
            assert ctx_client is bnd_ctx_client



import inspect
import pathlib
from collections.abc import Callable, Iterable, Iterator, Sequence
from pathlib import Path
from typing import Any

from escudeiro.data import data
from escudeiro.lazyfields import lazyfield

from .base import AutoDiscoveryHelper, PathConverter, StrOrPath, smart_import

type Validator = Callable[[Any], bool]


@data(frozen=False)
class RuntimeAutoDiscovery:
    """
    Facilitates runtime discovery and loading of modules based on a validation function.

    This class enables the discovery and loading of modules within a specified directory (`root`)
    or its subdirectories (`look_on`), leveraging runtime inspection techniques to identify and
    load modules that meet the criteria defined by the provided validator function.

    Attributes:
        _internal (AutoDiscoveryHelper): Helper instance for managing directory traversal and path conversion.
        _validator (Validator): Function to validate modules or objects based on runtime inspection.

    Methods:
        load() -> Iterator[tuple[str, Any]]:
            Lazily loads modules that pass the validation criteria defined by `_validator`.

        load_asdict() -> dict[str, Any]:
            Loads discovered modules into a dictionary format.

    Example Usage:
        validator = lambda obj: inspect.isclass(obj) and issubclass(obj, MyBaseClass)
        discovery = RuntimeAutoDiscovery(
            validator,
            root=pathlib.Path("/path/to/directory"),
            exclude=["excluded_module.py"],
            exts=["py"],
        )
        for name, obj in discovery.load():
            print(f"Loaded module: {name}")
    """

    _validator: Validator
    _internal: AutoDiscoveryHelper

    def __init__(
        self,
        validator: Validator,
        root: pathlib.Path,
        look_on: Path | None = None,
        exclude: Sequence[StrOrPath] = (),
        exts: Sequence[str] = (),
        converter: PathConverter = Path.as_posix,
        include: Sequence[StrOrPath] = (),
    ):
        """
        Iterates over the members of the module specified by the given path.

        Args:
            path (Path): The path to the module to be loaded.

        Yields:
            Iterable[tuple[str, Any]]: An iterable of tuples containing member names
            and their corresponding objects if they pass the validation.
        """
        self._validator = validator
        self._internal = AutoDiscoveryHelper(
            root,
            look_on,
            exclude,
            exts,
            converter,
            include,
        )

    def _itermod(self, path: Path) -> Iterable[tuple[str, Any]]:
        """
        Iterates over the members of the module specified by the given path.

        Args:
            path (Path): The path to the module to be loaded.

        Yields:
            Iterable[tuple[str, Any]]: An iterable of tuples containing member names
            and their corresponding objects if they pass the validation.
        """
        resolver = self._internal.get_resolver()
        mod = smart_import(resolver(path), resolver)

        for name, obj in inspect.getmembers(mod):
            if self._validator(obj):
                yield (name, obj)

    def load(self) -> Iterator[tuple[str, Any]]:
        """
        Lazily loads modules that pass the validation criteria.

        Returns:
            Iterator[tuple[str, Any]]: Iterator yielding tuples of module names and their corresponding objects.
        """
        for file in self._internal.iterdir(self._internal.target_path):
            for name, obj in self._itermod(file):
                if self._validator(obj):
                    yield name, obj

    @lazyfield
    def asdict(self) -> dict[str, Any]:
        """
        Loads discovered modules into a dictionary format.

        Returns:
            dict[str, Any]: Dictionary mapping module names to their corresponding objects.
        """
        return dict(self.load())

    def __iter__(self):
        yield from self.load()


def runtime_child_of(parent_class: type) -> Callable[[Any], bool]:
    """
    Returns a function that checks if an object is a subclass (or instance) of `parent_class`.

    Args:
        parent_class (type): The class or type to check against.

    Returns:
        Callable[[Any], bool]: Function that checks if an object is a subclass or instance of `parent_class`.
    """

    def check(obj: Any) -> bool:
        return isinstance(obj, type) and issubclass(obj, parent_class)

    return check


def runtime_instance_of(*bases: type) -> Callable[[Any], bool]:
    """
    Returns a function that validates instances of specified types in runtime.

    Args:
        *bases (type): Variable-length argument list of types to validate against.

    Returns:
        Callable[[Any], bool]: Function that validates instances of specified types.
    """

    def check(obj: Any) -> bool:
        return isinstance(obj, bases)

    return check


def runtime_contains_attr(attr_name: str) -> Callable[[Any], bool]:
    """
    Returns a function that checks if an object contains an attribute `attr_name`.

    Args:
        attr_name (str): The attribute name to check for.

    Returns:
        Callable[[Any], bool]: Function that checks if an object contains the specified attribute.
    """

    def check(obj: Any) -> bool:
        return hasattr(obj, attr_name)

    return check


def runtime_attr_with_value(
    attr_name: str, attr_value: Any
) -> Callable[[Any], bool]:
    """
    Returns a function that checks if an object has an attribute `attr_name` with a specific `attr_value`.

    Args:
        attr_name (str): The attribute name to check for.
        attr_value (Any): The expected value of the attribute.

    Returns:
        Callable[[Any], bool]: Function that checks if the object has the specified attribute with the specified value.
    """

    def check(obj: Any) -> bool:
        return hasattr(obj, attr_name) and getattr(obj, attr_name) == attr_value

    return check


def runtime_chain_validate_all(*validators: Validator) -> Validator:
    """Combines multiple validator functions into a single validator that checks if all validators pass."""

    def check(obj: Any) -> bool:
        return all(validator(obj) for validator in validators)

    return check


def runtime_chain_validate_any(*validators: Validator) -> Validator:
    """Combines multiple validator functions into a single validator that checks if any validator passes."""

    def check(obj: Any) -> bool:
        return any(validator(obj) for validator in validators)

    return check



# pyright: reportOptionalMemberAccess=false
import pytest

from escudeiro.data.converters.utils import asdict
from escudeiro.ds import VirtualFileTree
from escudeiro.exc import FailedFileOperation, InvalidPath


def test_create_dir():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        folder = inner_tree.create_dir("subfolder")
        assert folder.name == "subfolder"
        assert folder.children == []


def test_create_dir_existing_folder():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        folder1 = inner_tree.create_dir("subfolder")
        folder2 = inner_tree.create_dir("subfolder")
        # get_shallow returns a copy; we check name equality to confirm logical identity
        assert folder1.name == folder2.name


def test_create_dir_existing_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_file("file")
        with pytest.raises(InvalidPath):
            _ = inner_tree.create_dir("file")


def test_create_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        file = inner_tree.create_file("file")
        assert file.name == "file"
        assert file.content == b""


def test_create_file_existing_folder():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_dir("subfolder")
        with pytest.raises(InvalidPath):
            _ = inner_tree.create_file("subfolder")


def test_create_file_existing_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        file1 = inner_tree.create_file("file")
        file2 = inner_tree.create_file("file")
        assert file1.name == file2.name


def test_get_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_file("file")
        file = inner_tree.get_path("file")
        assert file.name == "file"


def test_get_file_nonexistent_folder():
    file_tree = VirtualFileTree.from_basename("root")
    with pytest.raises(InvalidPath):
        _ = file_tree.get_path("file", "invalid")


def test_get_file_nonexistent_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        with pytest.raises(InvalidPath):
            _ = inner_tree.get_path("file")


def test_get_dir():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_dir("folder")
        folder = inner_tree.get_path("folder")
        assert folder is not None
        assert folder.name == "folder"


def test_get_dir_nonexistent_folder():
    file_tree = VirtualFileTree.from_basename("root")
    with pytest.raises(InvalidPath):
        _ = file_tree.get_path("folder")


def test_create_py_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder"):
        file = file_tree.create_py_file("script")
        assert file.name == "script.py"


def test_init_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder"):
        file = file_tree.create_init_file()
        assert file.name == "__init__.py"


def test_from_virtual():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_dir("subfolder")
        _ = inner_tree.create_file("file")

        new_tree = VirtualFileTree.from_basename("new_root")
        new_tree.merge(inner_tree, "folder")
        new_inner_root = new_tree.root.get_shallow("folder")
        assert new_inner_root
        assert new_inner_root.name == "folder"

        old_inner_root = new_inner_root.get_shallow("folder")
        assert old_inner_root
        assert old_inner_root.get_shallow("subfolder")
        assert not old_inner_root.get_shallow("subfolder").is_file()
        assert old_inner_root.get_shallow("file")
        assert old_inner_root.get_shallow("file").is_file()


def test_from_virtual_existing_file():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_file("file")

    new_tree = VirtualFileTree.from_basename("file")
    with pytest.raises(InvalidPath):
        file_tree.merge(new_tree, "folder")


def test_from_virtual_foldername_conflict():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        _ = inner_tree.create_file("subfolder")

    new_tree = VirtualFileTree.from_basename("subfolder")
    with pytest.raises(InvalidPath):
        file_tree.merge(new_tree, "folder")


def test_virtual_context():
    file_tree = VirtualFileTree.from_basename("root")
    with file_tree.virtual_context("folder") as inner_tree:
        folder = inner_tree.create_dir("subfolder")
        file = inner_tree.create_file("file")
        text_file = inner_tree.create_text_file("text_file")
        text_file.append_content(b"Hello")
        assert folder.name == "subfolder"
        assert file.name == "file"
        assert inner_tree.root.get_shallow("subfolder")
        assert inner_tree.root.get_shallow("file")
    assert asdict(file_tree) == {
        "root": {
            "folder": {
                "subfolder": {},
                "file": file.content,
                "text_file": text_file.content,
            }
        },
    }


def test_virtual_context_exception():  # sourcery skip: raise-specific-error
    file_tree = VirtualFileTree.from_basename("root")
    with pytest.raises(FailedFileOperation):
        with file_tree.virtual_context("folder"):
            raise Exception("Something went wrong")

    assert file_tree.root.children == []



# pyright: reportPrivateUsage=false

import asyncio

from escudeiro.ds import TaskManager
from escudeiro.lazyfields import is_initialized


async def test_task_manager_start():
    manager = TaskManager()
    assert (
        not is_initialized(manager, "_worker_task")
        or manager._worker_task is None
    )
    _ = await manager.start()
    assert manager._worker_task is not None
    await manager.close()


async def test_task_manager_spawn():
    manager = await TaskManager().start()

    async def sample_task():
        await asyncio.sleep(0.1)

    manager.spawn(sample_task())
    assert not manager._pending_coro.empty()
    await manager.close()


async def test_task_manager_worker_execution():
    manager = await TaskManager().start()
    results = []

    async def sample_task():
        await asyncio.sleep(0.1)
        results.append(1)

    manager.spawn(sample_task())
    await asyncio.sleep(0.2)  # Allow time for execution
    assert results == [1]
    await manager.close()


async def test_task_manager_concurrency_limit():
    manager = await TaskManager(max_tasks=2).start()
    counter = 0
    lock = asyncio.Lock()

    async def sample_task():
        nonlocal counter
        async with lock:
            counter += 1
        await asyncio.sleep(0.2)
        async with lock:
            counter -= 1

    for _ in range(5):
        manager.spawn(sample_task())

    await asyncio.sleep(0.1)  # Allow some tasks to start
    assert counter <= 2  # max_tasks limit
    await manager.close()


async def test_task_manager_drain():
    manager = await TaskManager().start()
    completed = []

    async def sample_task():
        await asyncio.sleep(0.1)
        completed.append(1)

    manager.spawn(sample_task())
    await manager.drain()
    assert completed == [1]
    await manager.close()


async def test_task_manager_shutdown():
    manager = await TaskManager().start()
    running_task = asyncio.Event()

    async def sample_task():
        running_task.set()
        await asyncio.sleep(1)

    manager.spawn(sample_task())
    _ = await running_task.wait()
    await manager.close()
    assert not manager._running_tasks  # Ensure tasks are cleared


async def test_task_manager_context_manager():
    async with TaskManager() as manager:
        assert manager._worker_task is not None



import asyncio
import inspect
import uuid
from collections.abc import Awaitable, Callable
from datetime import datetime, timedelta, tzinfo
from typing import TYPE_CHECKING, Any

from escudeiro.data import data, field
from escudeiro.ds.manager import TaskManager
from escudeiro.escudeiro_pyrs import cronjob
from escudeiro.lazyfields import lazyfield
from escudeiro.misc import make_noop, now


@data(frozen=False)
class Task:
    id_: str
    func: Callable[[], Any]
    cron: cronjob.CronExpr | None
    next_run: datetime
    name: str | None = None
    is_running: bool = False
    remove_callback: Callable[[str], Awaitable[None]] = field(
        default=make_noop(asyncio=True)
    )

    @lazyfield
    def _count(self) -> int:
        """Counter tracking how many times the task has executed."""
        return 0

    @lazyfield
    def lock(self) -> asyncio.Lock:
        """Asyncio lock for thread-safe task state management."""
        return asyncio.Lock()

    async def run_count(self):
        """Get the number of times this task has executed.

        Returns:
            The execution count
        """
        return self._count

    async def set(self):
        """Mark the task as running and increment execution count."""
        async with self.lock:
            self.is_running = True
            self._count += 1

    async def reset(self):
        """Mark the task as not running."""
        async with self.lock:
            self.is_running = False

    async def status(self):
        """Check if the task is currently running.

        Returns:
            True if the task is running, False otherwise
        """
        async with self.lock:
            return self.is_running


@data
class CronInfo:
    """Container for cron expression scheduling information.

    Attributes:
        cronexpr: String representing the cron schedule
        timezone: Timezone for the schedule (defaults to local timezone)
    """

    cronexpr: str
    timezone: tzinfo = field(
        default_factory=lambda: datetime.now().astimezone().tzinfo
    )


@data(frozen=False)
class TaskScheduler:
    """Asynchronous task scheduler with cron-based and one-time execution support.

    Attributes:
        default_sleep_interval: Default sleep time (seconds) when no tasks exist
        log_info: Info logging callback
        log_warning: Warning logging callback
        log_exception: Error logging callback
    """

    default_sleep_interval: int = 1
    log_info: Callable[[str], Any] = make_noop()
    log_warning: Callable[[str], Any] = make_noop()
    log_exception: Callable[[str], Any] = make_noop()

    def __post_init__(self):
        _ = self._tasks, self._running, self.manager

    @lazyfield
    def _tasks(self) -> dict[str, Task]:
        """Dictionary mapping task IDs to Task instances."""
        return {}

    @lazyfield
    def _running(self):
        """Flag indicating if the scheduler is active."""
        return False

    @lazyfield
    def manager(self) -> TaskManager:
        """Task manager for handling background operations."""
        return TaskManager()

    @staticmethod
    def make_task_id() -> str:
        """Generate a unique task identifier.

        Returns:
            Hex string representing a UUID4
        """
        return uuid.uuid4().hex

    def schedule_task(
        self,
        func: Callable[[], Any],
        cron_expr: CronInfo | None,
        run_at: datetime | None,
        name: str | None = None,
        task_id: str | None = None,
        remove_callback: Callable[[str], Awaitable[None]] | None = None,
    ) -> str:
        """Schedule a new task for execution.

        Args:
            func: Callable to execute (sync or async)
            cron_expr: CronInfo for recurring tasks
            run_at: Specific datetime for one-time execution
            name: Optional task name (defaults to function name)
            task_id: Optional custom task ID
            remove_callback: Async callback when task is removed

        Returns:
            str: The task ID

        Raises:
            ValueError: If neither cron_expr nor run_at is provided,
                       or if both are provided
        """
        if not cron_expr and not run_at:
            raise ValueError("cron_expr or run_at must be provided.")

        if cron_expr is not None and run_at is not None:
            raise ValueError("Cannot specify both cron_expr and run_at")

        if run_at is not None:
            run_at = run_at.astimezone()
        task_id = task_id or self.make_task_id()
        cronexpr = cron_expr and cronjob.CronExpr(
            cron_expr.cronexpr, cron_expr.timezone
        )
        if cronexpr is not None:
            cronexpr.update()
            run_at = cronexpr.next_run

        if TYPE_CHECKING:
            assert run_at is not None

        task = Task(
            id_=task_id,
            func=func,
            cron=cronjob.CronExpr(cron_expr.cronexpr, cron_expr.timezone)
            if cron_expr
            else None,
            next_run=run_at,
            name=name or func.__name__,
            remove_callback=remove_callback or make_noop(asyncio=True),
        )

        self._tasks[task_id] = task
        self.log_info(f"Task '{task.name}' scheduled with ID {task_id}")
        return task_id

    def is_scheduled(self, task_id: str) -> bool:
        """Check if a task exists in the scheduler.

        Args:
            task_id: ID of the task to check

        Returns:
            bool: True if task exists, False otherwise
        """
        return task_id in self._tasks

    def get_task(self, task_id: str) -> Task | None:
        """Retrieve a task by its ID.

        Args:
            task_id: ID of the task to retrieve

        Returns:
            Task if found, None otherwise
        """
        return self._tasks.get(task_id)

    def task_ids(self) -> list[str]:
        """Get list of all scheduled task IDs.

        Returns:
            List of task ID strings
        """
        return list(self._tasks.keys())

    async def remove_task(
        self,
        task_id: str,
        callback: Callable[[str], Awaitable[None]],
    ) -> bool:
        """Remove a scheduled task.

        Args:
            task_id: ID of task to remove
            callback: Async callback to invoke after removal

        Returns:
            bool: True if task was removed, False if not found
        """
        if task_id not in self._tasks:
            return False
        task = self._tasks.pop(task_id)
        await callback(task_id)
        self.log_info(f"Taskt with ID {task_id} removed '{task.name}'")
        return True

    async def remove_by_name(
        self, name: str, callback: Callable[[str], Awaitable[None]]
    ) -> str | None:
        """Remove task(s) by name.

        Args:
            name: Name of task(s) to remove
            callback: Async callback to invoke after removal

        Returns:
            str: ID of removed task if found, None otherwise
        """
        for task_id, task in self._tasks.items():
            if task.name == name:
                _ = await self.remove_task(task_id, callback)
                return task_id
        return None

    async def _execute_task(self, task: Task) -> None:
        """Execute a task and handle post-execution logic.

        Args:
            task: Task instance to execute

        Note:
            Handles both synchronous and asynchronous functions,
            manages task state, and handles rescheduling/removal
        """
        if await task.status():
            self.log_warning(f"Task '{task.name}' is already running")
            return

        self.log_info(f"Starting task '{task.name}'")
        await task.set()
        try:
            if inspect.iscoroutinefunction(task.func):
                await task.func()
            else:
                await asyncio.to_thread(task.func)
            self.log_info(f"Task '{task.name}' finished.")

        except Exception as err:
            self.log_exception(f"Task '{task.name}' failed: {err}")

        finally:
            await task.reset()
            self._reschedule_task(task)

    def _reschedule_task(self, task: Task) -> None:
        """Reschedule a recurring task or queue for removal.

        Args:
            task: Task to reschedule

        Note:
            For cron-based tasks, calculates next run time.
            One-time tasks are queued for removal.
        """
        if task.cron:
            try:
                task.cron.update_after(
                    now().replace(second=0, microsecond=0)
                    + timedelta(minutes=1)
                )
                task.next_run = task.cron.next_run
            except Exception as e:
                self.log_exception(
                    f"Failed to reschedule task '{task.name}': {e}"
                )
                # Handle failed rescheduling
        else:
            self.manager.spawn(self.remove_task(task.id_, task.remove_callback))

    async def _scheduler_loop(self) -> None:
        """Main scheduling loop that executes due tasks.

        Continuously checks for and executes due tasks,
        sleeping between checks based on next scheduled run.
        """
        while self._running:
            timestamp = now()

            # Find and execute due tasks
            due_tasks = [
                task
                for task in self._tasks.values()
                if task.next_run <= timestamp
            ]

            # Check status only for tasks we'll actually run
            tasks_to_run = []
            for task in due_tasks:
                if not await task.status():
                    tasks_to_run.append(task)

            if tasks_to_run:
                await self._execute_tasks(tasks_to_run)

            await self._calculate_and_sleep()

    async def _calculate_and_sleep(self):
        """Calculate optimal sleep time before next task check.

        Sleeps until next scheduled task or default interval.
        """
        # Sleep for a short interval before next check
        if self._tasks:
            next_run_time = min(task.next_run for task in self._tasks.values())
            sleep_time = max(0, (next_run_time - now()).total_seconds())
        else:
            sleep_time = self.default_sleep_interval
        await asyncio.sleep(sleep_time)

    async def _execute_tasks(self, tasks: list[Task]) -> None:
        """Execute multiple tasks concurrently.

        Args:
            tasks: List of Task instances to execute

        Note:
            Gathers results and handles any exceptions
        """
        results = await asyncio.gather(
            *(self._execute_task(task) for task in tasks),
            return_exceptions=True,
        )
        for task, result in zip(tasks, results):
            if isinstance(result, Exception):
                self.log_exception(f"Task '{task.name}' failed {result}")

    async def setup(self) -> None:
        """Setup method to initialize the scheduler.

        Can be overridden for custom initialization logic.
        """
        _ = await self.manager.start()
        self.log_info("Task scheduler setup complete.")

    async def start(self) -> None:
        """Start the scheduler event loop."""
        if self._running:
            return

        self._running = True
        await self.setup()
        self.log_info("Task scheduler started.")
        await self._scheduler_loop()

    async def stop(self) -> None:
        """Stop the scheduler event loop."""
        self._running = False
        self.log_info("Task scheduler stopped")
        await self.manager.close()

    async def aclose(self):
        """Async cleanup method to stop scheduler and release resources.

        Alias for stop() for contextlib.aclosing.
        """
        await self.stop()



from pathlib import Path
from typing import final

import pytest
from hypothesis import given
from hypothesis.strategies import text

from escudeiro import config
from escudeiro.exc import InvalidCast, MissingName


@given(text(), text())
def test_config_call_success(key: str, value: str):
    """test config call returns valid string
    from mapping source when name exists"""
    mapping = config.EnvMapping({key: value})
    cfg = config.Config(mapping=mapping)

    val = cfg(key)

    assert isinstance(val, str)
    assert val == value


def test_config_call_name_not_found():
    """test config call raises `config.MissingName`
    if name does not exists in source mapping"""
    mapping = config.EnvMapping({})
    cfg = config.Config(mapping=mapping)

    with pytest.raises(MissingName):
        cfg("invalid")


def test_config_cast_returns_cast_type():
    """test config cast returns cast type
    when name exists"""

    @final
    class _Cast:
        def __init__(self, val: str) -> None:
            self.val = val

    key = "key"
    value = "val"
    mapping = config.EnvMapping({key: value})
    cfg = config.Config(mapping=mapping)

    casted = cfg(key, _Cast)

    assert isinstance(casted, _Cast)
    assert casted.val == value


def test_config_cast_fails_with_invalid_cast():
    """test config cast fails with invalid cast
    when cast raises TypeError or ValueError"""

    def _fail_cast(val: str):
        _ = val
        raise TypeError

    mapping = config.EnvMapping({"key": "value"})
    cfg = config.Config(mapping=mapping)

    with pytest.raises(InvalidCast):
        cfg("key", _fail_cast)


def test_env_mapping_raises_errors_correctly_on_read():
    mapping = config.EnvMapping({})
    mapping["my-name"] = "val"
    mapping["my-name"]

    with pytest.raises(KeyError):
        mapping["my-name"] = "error"
    with pytest.raises(KeyError):
        del mapping["my-name"]


def test_config_reads_from_env_file(tmp_path: Path):
    filename = tmp_path / ".envtestfile"
    with open(filename, "w") as buf:
        buf.writelines(
            map(
                lambda val: f"{val}\n",
                [
                    "TEST_VALUE_FOUND=world",  # Testing a normal key-value pair
                    "# FULLY_COMMENTED=error",  # Testing a comment (should be ignored)
                    "TEST_QUOTE_CAPTURE='123abc'",  # Testing single quotes around value
                    'TEST_QUOTE_CAPTURE_WITH_SPACES=" 321 "',  # Testing double quotes with spaces
                    "TEST_DOUBLE_QUOTE_CAPTURE=\"'123'\"",  # Testing double quotes around value
                    "TEST_COMMENT_IN_VALUE=abc # comment",  # Testing comment after value (should be ignored)
                    "TEST_QUOTED_COMMENT_CAPTURE='abc # comment'",  # Testing single quotes with quoted comment
                    "TEST_VALUE_WITH_HASH=abc#comment",  # Testing value with hash, comment should not be taken into consideration
                    "TEST_QUOTE_WITH_COMMENT_CAPTURE='abc' # \"comment\"",  # Testing single quotes with inline comment
                    "TEST_QUOTE_CAPTURE_WITH_MIXED_COMMENTS='abc # comment\"",  # Testing single quotes with mixed comment format
                    "TEST_EMPTY_VALUE=",  # Testing empty value
                ],
            )
        )
    cfg = config.Config(filename)
    assert cfg("TEST_VALUE_FOUND") == "world"
    assert cfg("TEST_QUOTE_CAPTURE") == "123abc"
    assert cfg("TEST_QUOTE_CAPTURE_WITH_SPACES") == " 321 "
    assert cfg("TEST_DOUBLE_QUOTE_CAPTURE") == "'123'"
    assert cfg("TEST_COMMENT_IN_VALUE") == "abc"
    assert cfg("TEST_QUOTED_COMMENT_CAPTURE") == "abc # comment"
    assert cfg("TEST_VALUE_WITH_HASH") == "abc#comment"
    assert cfg("TEST_QUOTE_WITH_COMMENT_CAPTURE") == "abc"
    assert cfg("TEST_QUOTE_CAPTURE_WITH_MIXED_COMMENTS") == "'abc"
    assert cfg("TEST_EMPTY_VALUE") == ""

    with pytest.raises(MissingName):
        cfg("FULLY_COMMENTED")



"""Utilities for working with iterables and sequences.

This module provides functions for common iteration patterns, sequence transformations,
and advanced iterator operations with strong type safety.
"""

from __future__ import annotations

import itertools
from collections import defaultdict
from collections.abc import (
    AsyncIterable,
    Callable,
    Collection,
    Coroutine,
    Hashable,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
)
from typing import TYPE_CHECKING, Any, cast, overload

from escudeiro.misc.functions import as_async, safe_cast


def moving_window[T](
    iterable: Iterable[T],
    window_size: int,
    cast: Callable[[Iterator[T]], Sequence[T]] = tuple,
) -> Iterable[Sequence[T]]:
    """Returns an iterator yielding non-overlapping windows from an iterable.

    Creates a sliding window of fixed size that moves through the iterable
    one element at a time.

    Args:
        iterable: The input iterable to create windows from.
        window_size: The size of each window.
        cast: A function to convert each window iterator to a sequence type.
             Defaults to tuple.

    Returns:
        An iterable of windows, where each window is a sequence of length
        window_size or smaller (for the final window if iterable length is
        not a multiple of window_size).

    Examples:
        ```python
        # Get sliding windows of size 3 from a list
        list(moving_window([1, 2, 3, 4, 5], 3))
        # Output: [(1, 2, 3), (4, 5)]

        # Use different output sequence type
        list(moving_window([1, 2, 3, 4, 5], 2, list))
        # Output: [[1, 2], [3, 4], [5]]
        ```
    """
    iterator = iter(iterable)

    while True:
        window = cast(itertools.islice(iterator, window_size))
        if not window:
            break
        yield window


@overload
def aislice[T](
    iterable: AsyncIterable[T],
    stop: int | None = ...,
    /,
) -> AsyncIterable[T]: ...


@overload
def aislice[T](
    iterable: AsyncIterable[T],
    start: int | None,
    stop: int | None,
    step: int | None = ...,
    /,
) -> AsyncIterable[T]: ...


async def aislice[T](
    iterable: AsyncIterable[T], *args: int | None
) -> AsyncIterable[T]:
    """Async slice iterator.

    Examples:
        >>> async for item in aislice(stream, 5):  # First 5 items
        ...     process(item)
        >>> async for item in aislice(stream, 2, 10, 2):  # Items 2,4,6,8
        ...     process(item)
    """
    s = slice(*args)
    start, stop, step = s.start or 0, s.stop, s.step or 1

    if not all(isinstance(x, int) for x in (start, step) if x is not None):
        raise TypeError("Slice indices must be integers")
    if any(x < 0 for x in (start, stop, step) if x is not None):
        raise ValueError("Negative indices not supported")
    if s.step == 0:
        raise ValueError("Step cannot be zero")
    if stop is not None and start >= stop:
        return  # Early exit for empty slices
    it = aiter(iterable)

    # Skip initial elements
    for _ in range(start):
        try:
            _ = await anext(it)
        except StopAsyncIteration:
            return

    # Yield remaining elements
    while stop is None or start < stop:
        try:
            yield await anext(it)
            start += step
            # Skip step-1 elements
            for _ in range(step - 1):
                _ = await anext(it)
        except StopAsyncIteration:
            break


async def amoving_window[T](
    iterable: AsyncIterable[T],
    window_size: int,
) -> AsyncIterable[Sequence[T]]:
    """Returns an iterator yielding non-overlapping windows from an iterable.

    Creates a sliding window of fixed size that moves through the iterable
    one element at a time.

    Args:
        iterable: The input iterable to create windows from.
        window_size: The size of each window.

    Returns:
        An iterable of windows, where each window is a sequence of length
        window_size or smaller (for the final window if iterable length is
        not a multiple of window_size).
    """
    iterator = aiter(iterable)

    while True:
        window = [item async for item in aislice(iterator, window_size)]
        if not window:
            break
        yield window


async def amap[V, T](
    predicate: Callable[[V], T] | Callable[[V], Coroutine[Any, Any, T]],
    iterable: AsyncIterable[V],
) -> AsyncIterable[T]:
    predicate = cast(Callable[[V], Coroutine[Any, Any, T]], as_async(predicate))
    async for item in iterable:
        yield await predicate(item)


async def afilter[V](
    predicate: Callable[[V], bool] | Callable[[V], Coroutine[Any, Any, bool]],
    iterable: AsyncIterable[V],
) -> AsyncIterable[V]:
    predicate = cast(
        Callable[[V], Coroutine[Any, Any, bool]], as_async(predicate)
    )
    async for item in iterable:
        if await predicate(item):
            yield item


_reduce_sentinel = object()


@overload
async def areduce[T](
    predicate: Callable[[T, T], T] | Callable[[T, T], Coroutine[Any, Any, T]],
    iterable: AsyncIterable[T],
    initial: object = _reduce_sentinel,
) -> T: ...


@overload
async def areduce[T](
    predicate: Callable[[T, T], T] | Callable[[T, T], Coroutine[Any, Any, T]],
    iterable: AsyncIterable[T],
    initial: T,
) -> T: ...


@overload
async def areduce[V, T](
    predicate: Callable[[V, T], T] | Callable[[V, T], Coroutine[Any, Any, T]],
    iterable: AsyncIterable[T],
    initial: V,
) -> T: ...


async def areduce[V, T](
    predicate: Callable[[T, V], T]
    | Callable[[T, V], Coroutine[Any, Any, T]]
    | Callable[[V, V], V]
    | Callable[[V, V], Coroutine[Any, Any, V]],
    iterable: AsyncIterable[V],
    initial: T | V | object = _reduce_sentinel,
) -> T | V:
    iterable = aiter(iterable)
    predicate = cast(
        Callable[[T | V, V], Coroutine[Any, Any, T]], as_async(predicate)
    )

    if initial is _reduce_sentinel:
        try:
            value = await anext(iterable)
        except StopAsyncIteration:
            raise TypeError(
                "reduce() of empty iterable with no initial value"
            ) from None
    else:
        value = cast(T | V, initial)
    async for item in iterable:
        value = await predicate(value, item)
    return value


async def aenumerate[T](
    iterable: AsyncIterable[T], start: int = 0
) -> AsyncIterable[tuple[int, T]]:
    """Return an async iterator that yields tuples of (index, value)."""
    index = start
    async for item in iterable:
        yield index, item
        index += 1


async def aany[T](
    iterable: AsyncIterable[T],
    predicate: Callable[[T], bool]
    | Callable[[T], Coroutine[Any, Any, bool]] = bool,
) -> bool:
    """Return True if any item in the async iterable satisfies the predicate."""
    return (await anext_or(afilter(predicate, iterable))) is not None


async def aall[T](
    iterable: AsyncIterable[T],
    predicate: Callable[[T], bool]
    | Callable[[T], Coroutine[Any, Any, bool]] = bool,
) -> bool:
    """Return True if all items in the async iterable satisfy the predicate."""

    async def _not_predicate(x: T) -> bool:
        return not await as_async(predicate)(x)

    return not await aany(iterable, _not_predicate)


async def acarrymap[U, T](
    predicate: Callable[[T], Coroutine[Any, Any, U]], iterable: AsyncIterable[T]
) -> AsyncIterable[tuple[U, T]]:
    """Return an async iterator that yields tuples of (result, arg)
    where the result is the result of applying the predicate to the arg."""
    async for arg in iterable:
        yield await predicate(arg), arg


async def astarmap[*Ts, U](
    predicate: Callable[[*Ts], Coroutine[Any, Any, U]],
    iterable: AsyncIterable[tuple[*Ts]],
) -> AsyncIterable[U]:
    """Return an async iterator that yields the result
    where the result is the result of applying the predicate to the args."""
    async for args in iterable:
        yield await predicate(*args)


async def acarrystarmap[*Ts, U](
    predicate: Callable[[*Ts], Coroutine[Any, Any, U]],
    iterable: AsyncIterable[tuple[*Ts]],
) -> AsyncIterable[tuple[U, tuple[*Ts]]]:
    """Return an async iterator that yields tuples of (result, args)
    where the result is the result of applying the predicate to the args."""
    async for args in iterable:
        yield await predicate(*args), args


def flatten(sequence: Sequence[Any]) -> Sequence[Any]:
    """Flattens nested sequences into a single-level sequence.

    Recursively flattens nested sequences (lists, tuples, etc.) while
    preserving the outer sequence type. Strings and bytes objects are
    treated as atomic units and not flattened.

    Args:
        sequence: The nested sequence to flatten.

    Returns:
        A flattened sequence of the same type as the input sequence.
        If the original type cannot be preserved, returns a list.

    Examples:
        ```python
        flatten([1, [2, 3], [4, [5, 6]]])
        # Output: [1, 2, 3, 4, 5, 6]

        flatten((1, [2, 3], (4, 5)))
        # Output: (1, 2, 3, 4, 5)
        ```
    """
    flattened: list[Any] = []
    stack: list[tuple[Sequence[Any], int]] = [(sequence, 0)]

    while stack:
        curseq, index = stack.pop()
        while index < len(curseq):
            item = curseq[index]
            index += 1
            if isinstance(item, Sequence) and not isinstance(item, str | bytes):
                stack.append((curseq, index))
                curseq, index = item, 0
            else:
                flattened.append(item)

    return safe_cast(type(sequence), flattened, Exception, default=flattened)


SequenceTypes = list | dict | set | tuple


def exclude_none[SequenceT: SequenceTypes](sequence: SequenceT) -> SequenceT:
    """Recursively filters out `None` values from sequences and their nested elements.

    Args:
        sequence: The sequence to filter. Must be a dict, list, set, or tuple.

    Returns:
        Filtered sequence with `None` values removed, preserving the original
        sequence type.

    Notes:
        - Tuple typing will not be preserved due to Python's immutability of tuples.
        - For dictionaries, only values are checked for None; keys are always preserved.

    Examples:
        ```python
        # Remove None from list
        exclude_none([1, None, 2, [3, None, 4]])
        # Output: [1, 2, [3, 4]]

        # Remove None from dictionary
        exclude_none({"a": 1, "b": None, "c": {"d": None, "e": 2}})
        # Output: {"a": 1, "c": {"e": 2}}
        ```
    """
    outer_acc: SequenceT = (
        type(sequence)() if not isinstance(sequence, tuple) else []
    )
    stack: list[tuple[SequenceT, SequenceT]] = [(sequence, outer_acc)]

    while stack:
        curr, acc = stack.pop()

        if isinstance(curr, dict):
            if TYPE_CHECKING:
                acc = cast(dict, acc)
            for key, value in curr.items():
                if value is None:
                    continue
                if isinstance(value, SequenceTypes):
                    new_acc = (
                        type(value)() if not isinstance(value, tuple) else []
                    )
                    acc[key] = new_acc
                    stack.append((value, new_acc))  # pyright: ignore[reportArgumentType]
                else:
                    acc[key] = value
        else:
            temp_acc = []
            for item in curr:
                if item is None:
                    continue
                if isinstance(item, SequenceTypes):
                    new_acc = (
                        type(item)() if not isinstance(item, tuple) else []
                    )
                    temp_acc.append(new_acc)
                    stack.append((item, new_acc))  # pyright: ignore[reportArgumentType]
                else:
                    temp_acc.append(item)
            if isinstance(curr, list | tuple):
                _ = acc.extend(temp_acc)  # pyright: ignore[reportAttributeAccessIssue]
            else:
                _ = acc.update(temp_acc)  # pyright: ignore[reportAttributeAccessIssue]

    return outer_acc


def next_or[T, D](iterable: Iterable[T], default: D = None) -> T | D:
    """Returns the first element from an iterable or a default value if empty.

    A convenience wrapper around the built-in `next` function with a default value.

    Args:
        iterable: The iterable to get the first element from.
        default: The value to return if the iterable is empty. Defaults to None.

    Returns:
        The first element of the iterable or the default value if the iterable is empty.

    Examples:
        ```python
        next_or([1, 2, 3])  # Returns 1
        next_or([], "empty")  # Returns "empty"
        ```
    """
    return next(iter(iterable), default)


async def anext_or[T, D](
    iterable: AsyncIterable[T], default: D = None
) -> T | D:
    """Returns the first element from an iterable or a default value if empty.

    A convenience wrapper around the built-in `anext` function with a default value.

    Args:
        iterable: The iterable to get the first element from.
        default: The value to return if the iterable is empty. Defaults to None.

    Returns:
        The first element of the iterable or the default value if the iterable is empty.

    Examples:
        ```python
        await anext_or([1, 2, 3])  # Returns 1
        await anext_or([], "empty")  # Returns "empty"
        ```
    """
    return await anext(aiter(iterable), default)


def carrymap[T, U](
    predicate: Callable[[T], U], iterable: Iterable[T]
) -> Iterable[tuple[U, T]]:
    """Maps elements with a function while preserving the original values.

    Applies a function to each element in an iterable and yields tuples of
    (result, original), where result is the transformed value and original
    is the original input element.

    Args:
        predicate: A function to apply to each element in the iterable.
        iterable: The input iterable.

    Returns:
        An iterator yielding tuples of (transformed_value, original_value).

    Examples:
        ```python
        # Transform elements while keeping originals
        list(carrymap(str.upper, ["a", "b", "c"]))
        # Output: [("A", "a"), ("B", "b"), ("C", "c")]

        # Calculate lengths while preserving strings
        list(carrymap(len, ["apple", "banana", "cherry"]))
        # Output: [(5, "apple"), (6, "banana"), (6, "cherry")]
        ```
    """
    for arg in iterable:
        yield predicate(arg), arg


def filter_isinstance[T](
    bases: type[T] | tuple[type[T], ...], iterable: Iterable[Any]
) -> filter[T]:
    """Filters an iterable to include only instances of specified types.

    Creates a filter object that yields only elements from the iterable
    that are instances of the given type or types.

    Args:
        bases: A type or tuple of types to check against.
        iterable: The iterable to filter.

    Returns:
        A filter object yielding only elements that are instances of the specified types.

    Examples:
        ```python
        # Filter to keep only strings
        list(filter_isinstance(str, [1, "hello", 2, "world"]))
        # Output: ["hello", "world"]

        # Filter to keep strings or numbers
        list(filter_isinstance((str, int), [1, "hello", [], 2, {}, "world"]))
        # Output: [1, "hello", 2, "world"]
        ```
    """

    def _predicate(item: Any) -> bool:
        return isinstance(item, bases)

    return filter(_predicate, iterable)


def filter_issubclass[T](bases: type[T], iterable: Iterable[Any]) -> filter[T]:
    """Filters an iterable to include only types that are subclasses of specified types.

    Creates a filter object that yields only elements from the iterable
    that are types (classes) and are subclasses of the given type or types.

    Args:
        bases: A type or tuple of types to check against.
        iterable: The iterable to filter. Elements should be types (classes).

    Returns:
        A filter object yielding only types that are subclasses of the specified types.

    Examples:
        ```python
        # Filter to keep only exception subclasses
        classes = [ValueError, str, TypeError, list, OSError]
        list(filter_issubclass(Exception, classes))
        # Output: [ValueError, TypeError, OSError]

        # Filter to keep only sequence subclasses
        list(filter_issubclass(Sequence, [list, dict, tuple, set]))
        # Output: [list, tuple]
        ```
    """

    def _predicate(item: Any) -> bool:
        return isinstance(item, type) and issubclass(item, bases)

    return filter(_predicate, iterable)


def invert_dict[K: Hashable, V: Hashable](
    mapping: Mapping[K, V],
) -> Mapping[V, K]:
    """
    Inverts the given mapping by swapping its keys and values.

    This function creates a new mapping where each value from the input
    becomes a key, and each key becomes the corresponding value. It is
    intended for mappings with unique values to avoid data loss. If the
    input mapping contains duplicate values, only one of the corresponding
    keys will be retained in the inverted mapping.

    Args:
        mapping: A mapping (such as a dict) where both keys and values are hashable.

    Returns:
        A new mapping with keys and values swapped.

    Raises:
        TypeError: If the values in the original mapping are not hashable,
                   since they are used as keys in the inverted mapping.

    Examples:
        >>> invert_dict({'a': 1, 'b': 2})
        {1: 'a', 2: 'b'}

        >>> invert_dict({1: 'x', 2: 'x'})
        { 'x': 2 }  # Note: The key 1 is overwritten by 2 because values were duplicated.
    """
    return {value: key for key, value in mapping.items()}


def group_values[K, V](
    vals: Collection[dict[K, V]], group_by_key: K
) -> dict[K, list[dict[K, V]]]:
    """
    Groups a collection of dictionaries by a specified key.

    This function iterates over a collection of dictionaries and aggregates them into a new
    dictionary. Each key in the returned dictionary corresponds to a unique value found under
    the specified group_by_key in the input dictionaries, and the associated value is a list
    of dictionaries that have that same key value.

    Args:
        vals: A collection of dictionaries. Each dictionary is expected to contain the key
              specified by group_by_key.
        group_by_key: The key used to group the dictionaries. The value for this key must be
                      hashable and present in every dictionary within the collection.

    Returns:
        A dictionary mapping each unique value (found under group_by_key) to a list of dictionaries
        that share that value.

    Raises:
        KeyError: If any dictionary in vals does not contain the specified group_by_key.

    Examples:
        >>> data = [
        ...     {'category': 'A', 'value': 1},
        ...     {'category': 'B', 'value': 2},
        ...     {'category': 'A', 'value': 3},
        ... ]
        >>> group_values(data, 'category')
        {'A': [{'category': 'A', 'value': 1}, {'category': 'A', 'value': 3}],
         'B': [{'category': 'B', 'value': 2}]}
    """
    result = defaultdict(list)
    for item in vals:
        result[item[group_by_key]].append(item)
    return dict(result)



from collections.abc import Mapping
from typing import Self

from escudeiro.escudeiro_pyrs import url
from escudeiro.url.fragment import Fragment
from escudeiro.url.mixins import Wrapped
from escudeiro.url.netloc import Netloc
from escudeiro.url.path import Path
from escudeiro.url.query import Query


class URL(Wrapped[url.URL]):
    def __init__(self, val: str) -> None:
        super().__init__(url.URL(val))

    @property
    def scheme(self) -> str:
        return self.internal.scheme

    @scheme.setter
    def scheme(self, val: str) -> None:
        self.internal.scheme = val

    @property
    def netloc(self) -> Netloc:
        return Netloc.from_internal(self.internal.netloc)

    @netloc.setter
    def netloc(self, val: Netloc) -> None:
        self.internal.netloc = val.internal

    @property
    def path(self) -> Path:
        return Path.from_internal(self.internal.path)

    @path.setter
    def path(self, val: Path) -> None:
        self.internal.path = val.internal

    @property
    def query(self) -> Query:
        return Query.from_internal(self.internal.query)

    @query.setter
    def query(self, val: Query) -> None:
        self.internal.query = val.internal

    @property
    def fragment(self) -> Fragment:
        return Fragment.from_internal(self.internal.fragment)

    @fragment.setter
    def fragment(self, val: Fragment) -> None:
        self.internal.fragment = val.internal

    def copy(self) -> Self:
        new_instance = object.__new__(type(self))
        new_instance.internal = self.internal.copy()
        return new_instance

    def add(
        self,
        path: str | None = None,
        query: Mapping[str, str] | None = None,
        fragment: str | None = None,
        netloc: str | None = None,
        netloc_obj: Netloc | None = None,
        scheme: str | None = None,
    ) -> Self:
        self.internal.add(
            path,
            query,
            fragment,
            netloc,
            netloc_obj and netloc_obj.internal,
            scheme,
        )
        return self

    def set(
        self,
        path: str | None = None,
        query: Mapping[str, str] | None = None,
        fragment: str | None = None,
        netloc: str | None = None,
        netloc_obj: Netloc | None = None,
        scheme: str | None = None,
    ) -> Self:
        self.internal.set(
            path,
            query,
            fragment,
            netloc,
            netloc_obj and netloc_obj.internal,
            scheme,
        )
        return self

    @classmethod
    def from_netloc(
        cls,
        netloc: Netloc | None = None,
        username: str | None = None,
        password: str | None = None,
        host: str | None = None,
        port: int | None = None,
    ) -> Self:
        return cls.from_internal(
            url.URL.from_netloc(
                netloc and netloc.internal, username, password, host, port
            )
        )

    @classmethod
    def from_args(
        cls,
        path: str | None = None,
        query: Mapping[str, str] | None = None,
        fragment: str | None = None,
        netloc: str | None = None,
        netloc_obj: Netloc | None = None,
        scheme: str | None = None,
    ) -> Self:
        return cls.from_internal(
            url.URL.from_args(
                path,
                query,
                fragment,
                netloc,
                netloc_obj and netloc_obj.internal,
                scheme,
            )
        )

    def encode(self, append_empty_equal: bool = True) -> str:
        return self.internal.encode(append_empty_equal)

    def copy_add(
        self,
        path: str | None = None,
        query: Mapping[str, str] | None = None,
        fragment: str | None = None,
        netloc: str | None = None,
        netloc_obj: Netloc | None = None,
        scheme: str | None = None,
    ) -> Self:
        return self.copy().add(
            path,
            query,
            fragment,
            netloc,
            netloc_obj,
            scheme,
        )

    def copy_set(
        self,
        path: str | None = None,
        query: Mapping[str, str] | None = None,
        fragment: str | None = None,
        netloc: str | None = None,
        netloc_obj: Netloc | None = None,
        scheme: str | None = None,
    ) -> Self:
        return self.copy().set(
            path,
            query,
            fragment,
            netloc,
            netloc_obj,
            scheme,
        )



from enum import Enum
from pathlib import Path
from typing import Any, Literal

import pytest

from escudeiro.config import Config, EnvMapping
from escudeiro.config.core.utils import (
    boolean_cast,
    literal_cast,
    none_is_missing,
    valid_path,
)
from escudeiro.exc import InvalidCast, InvalidEnv, MissingName
from escudeiro.misc.pathx import is_valid_path
from escudeiro.misc.strings import squote


def test_boolean_returns_valid_bool():
    mapping = EnvMapping(
        {"first": "true", "second": "False", "third": "1", "fourth": "0"}
    )
    cfg = Config(mapping=mapping)

    assert cfg("first", boolean_cast)
    assert not cfg("second", boolean_cast)
    assert cfg("third", boolean_cast)
    assert not cfg("fourth", boolean_cast)
    assert boolean_cast.strict(True) is True
    assert boolean_cast.strict(False) is False


def test_boolean_raises_invalid_cast():
    """test boolean raises invalid cast if
    no boolean definition matches"""
    mapping = EnvMapping({"key": "value"})
    cfg = Config(mapping=mapping)

    with pytest.raises(InvalidCast):
        _ = cfg("key", boolean_cast.strict)


def test_valid_path_returns_path_object(tmp_path: Path):
    filepath = tmp_path / "file.txt"
    filepath.touch()
    mapping = EnvMapping({"key": filepath.as_posix()})
    cfg = Config(mapping=mapping)

    val = cfg("key", valid_path)

    assert isinstance(val, Path)
    assert val == filepath


def test_valid_path_raises_file_not_found_error():
    """test valid_path raises FileNotFoundError
    if the path does not exist."""
    mapping = EnvMapping({"key": "./non_existent_file.txt"})
    cfg = Config(mapping=mapping)
    valpath = Path("./non_existent_file.txt")

    with pytest.raises(InvalidCast) as exc_info:
        _ = cfg("key", valid_path)

    assert isinstance(exc_info.value.__cause__, InvalidCast)
    assert exc_info.value.__cause__.args == (
        f"result {valpath!s} does not satisfy the rule {squote(is_valid_path.__name__)}.",
        valpath,
    )


def test_literal_cast_returns_valid_cast():
    class Test(Enum):
        VALUE = "value"

    literal_type = Literal[1, "other", b"another", Test.VALUE, None, False]
    caster = literal_cast(literal_type)
    mapping = EnvMapping(
        {
            "first": "other",
            "second": "another",
            "third": "1",
            "fourth": "value",
            "fifth": "null",
            "sixth": "false",
            "seventh": "invalid",
        }
    )
    cfg = Config(mapping=mapping)

    assert (
        cfg("first", caster),
        cfg("second", caster),
        cfg("third", caster),
        cfg("fourth", caster),
        cfg("fifth", caster),
        cfg("sixth", caster),
    ) == (
        "other",
        b"another",
        1,
        Test.VALUE,
        None,
        False,
    )

    with pytest.raises(InvalidCast) as exc_info:
        cfg("seventh", caster)

    assert exc_info.value.__cause__ is not None
    assert exc_info.value.__cause__.args == (
        "Value received does not match any argument from literal",
        literal_type.__args__,
    )


def test_none_is_missing():
    mapping = EnvMapping({"key": "null"})
    cfg = Config(mapping=mapping)

    with pytest.raises(MissingName):
        _ = cfg("key", none_is_missing(boolean_cast.optional))



import asyncio
import logging
from collections.abc import Callable, Coroutine
from functools import partial, wraps
from typing import Any, ParamSpec, TypeVar

from escudeiro.data import data, field

T = TypeVar("T")
P = ParamSpec("P")


def _default_on_err(_: Exception):
    """Logs an exception that occurs during CircuitBreaker execution.

    Args:
        exc (Exception): The exception that was raised.
    """
    logging.exception("Failed execution during CircuitBreaker execution")


DEFAULT_DELAY = 5  # seconds


@data(frozen=False)
class CircuitBreaker:
    """Implements a circuit breaker pattern for handling failures in async operations.

    Attributes:
        freeze_function (Callable[[], Coroutine]): The function to execute during the frozen state.
        on_error (Callable[[Exception], None]): The callback to execute when an error occurs.
    """

    freeze_function: Callable[[], Coroutine] = partial(
        asyncio.sleep, DEFAULT_DELAY
    )
    on_error: Callable[[Exception], None] = _default_on_err
    _lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    _frozen: bool = field(default=False)
    _freeze_future: asyncio.Future | None = field(default=None)

    async def execute(
        self,
        func: Callable[P, Coroutine[Any, Any, T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        """Executes the provided function with circuit breaker logic.

        If the circuit breaker is frozen, waits for it to unfreeze before retrying execution.
        If an exception occurs, triggers the freezing mechanism.

        Args:
            func (Callable[P, Coroutine[Any, Any, T]]): The async function to execute.
            *args (P.args): Positional arguments to pass to the function.
            **kwargs (P.kwargs): Keyword arguments to pass to the function.

        Returns:
            T: The result of the function execution.

        Raises:
            Any exception that occurs during execution after the circuit breaker is unfrozen.
        """
        if not self._frozen:
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                self.on_error(e)
                async with self._lock:
                    if not self._frozen:
                        await self._freeze(e)

        if self._freeze_future is not None:
            await self._freeze_future

        return await func(*args, **kwargs)

    @property
    def is_frozen(self):
        return self._frozen

    async def _freeze(self, __exc__: Exception) -> None:
        """Freezes the circuit breaker for a predefined duration.

        Args:
            error (Exception): The exception that caused the freeze.
        """
        self._frozen = True
        loop = asyncio.get_running_loop()
        frozen_future = loop.create_future()
        self._freeze_future = frozen_future

        async def unfreeze():
            try:
                await self.freeze_function()
                self._frozen = False
                frozen_future.set_result(None)
            except Exception as e:
                frozen_future.set_exception(e)

        _ = asyncio.create_task(unfreeze())


def with_circuit_breaker(circuit_breaker: CircuitBreaker):
    """Decorator that applies a CircuitBreaker to an async function.

    Args:
        circuit_breaker (CircuitBreaker): The circuit breaker instance to use.

    Returns:
        Callable[P, Coroutine[Any, Any, T]]: A wrapped function that applies circuit breaker logic.
    """

    def decorator(
        func: Callable[P, Coroutine[Any, Any, T]],
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            """Wrapper function that executes the given function within the circuit breaker."""
            return await circuit_breaker.execute(func, *args, **kwargs)

        return wrapper

    return decorator



from pathlib import Path

import pytest

from escudeiro.config import DotFile, Env, EnvConfig, EnvMapping
from escudeiro.exc import InvalidCast, MissingName

PARENT_DIR = curpath = Path(__file__).resolve().parent


def test_envconfig_returns_file_vals_if_in_expected_env():
    """EnvConfig should consider file_vals
    if env matches consider_file_on_env"""
    environ = EnvMapping({"CONFIG_ENV": Env.LOCAL.name})
    config = EnvConfig(
        DotFile(PARENT_DIR / ".envconfig", Env.LOCAL),
        mapping=environ,
    )

    assert config.env is Env.LOCAL
    assert config("NAME") == "teste"


def test_envconfig_does_not_find_value_if_not_in_expected_env():
    environ = EnvMapping({"CONFIG_ENV": Env.DEV.name})
    config = EnvConfig(
        DotFile(PARENT_DIR / ".envconfig", Env.LOCAL),
        mapping=environ,
    )

    with pytest.raises(MissingName):
        config("NAME")


def test_envconfig_raises_missing_name_if_no_env_is_found():
    with pytest.raises(MissingName):
        _ = EnvConfig(DotFile(PARENT_DIR / ".envconfig", Env.LOCAL))


def test_envconfig_raises_invalid_cast_if_env_val_is_invalid():
    environ = EnvMapping({"CONFIG_ENV": "invalid"})

    with pytest.raises(InvalidCast):
        _ = EnvConfig(
            DotFile(PARENT_DIR / ".envconfig", Env.LOCAL),
            mapping=environ,
        )


def test_envconfig_opens_config_dotfile_if_passed():
    environ = EnvMapping({"CONFIG_DOTFILE": "tests/config/.envconfig"})

    env_config = EnvConfig(DotFile("invalid", Env.LOCAL), mapping=environ)
    assert env_config.dotfile
    assert env_config.dotfile.filename == Path("tests/config/.envconfig")
    assert env_config.dotfile.env == Env.ALWAYS
    assert env_config("NAME") == "teste"


def test_env_config_uses_default_env_if_none_is_provided():
    environ = EnvMapping({"CONFIG_ENV": Env.DEV.name})
    config = EnvConfig(
        DotFile(PARENT_DIR / ".envconfig", Env.LOCAL), default_env=Env.LOCAL
    )
    with_env = EnvConfig(
        DotFile(PARENT_DIR / ".envconfig", Env.DEV),
        mapping=environ,
        default_env=Env.LOCAL,
    )

    assert config.env is Env.LOCAL
    assert with_env.env is Env.DEV
    assert (config("NAME"), with_env("NAME")) == ("teste", "teste")


def test_env_config_supports_envless_environment():
    config = EnvConfig(
        DotFile(PARENT_DIR / ".envconfig", Env.LOCAL),
        mapping=EnvMapping({}),
        strict=False,
    )

    assert config.env is None



from typing import Any

from escudeiro.config import DotFile, Env, EnvConfig, EnvMapping


def test_env_cast_default():
    environ = EnvMapping({"MY_ENV_VAR": Env.LOCAL.name})
    dotfile = DotFile(filename="mydotfile", env=Env.TEST)
    config = EnvConfig(dotfile, env_var="MY_ENV_VAR", mapping=environ)

    # Define your test cases for the default env_cast
    test_cases = [
        ("local", Env.LOCAL),
        ("test", Env.TEST),
        ("dev", Env.DEV),
        ("qa", Env.QA),
        ("prd", Env.PRD),
    ]

    for input_value, expected_env in test_cases:
        result_env = config.env_cast(input_value)
        assert result_env == expected_env


def test_env_cast_custom_handler():
    def custom_env_cast(input_value: Any):
        # Custom logic for casting environment names to Env enum values
        if input_value == "custom":
            return Env.DEV
        return Env.PRD  # Default if custom value not recognized

    environ = EnvMapping({"MY_ENV_VAR": Env.LOCAL.name})
    dotfile = DotFile(filename="mydotfile", env=Env.TEST)
    config = EnvConfig(
        dotfile, env_var="MY_ENV_VAR", mapping=environ, env_cast=custom_env_cast
    )

    # Define your test cases for the custom env_cast
    test_cases = [
        ("local", Env.PRD),
        ("test", Env.PRD),
        ("custom", Env.DEV),
        (
            "unknown",
            Env.PRD,
        ),  # Custom handler returns default for unknown values
    ]

    for input_value, expected_env in test_cases:
        result_env = config.env_cast(input_value)
        assert result_env == expected_env



from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import (
    Any,
    ForwardRef,
    NamedTuple,
    Protocol,
    Self,
    TypedDict,
    override,
)


class DisassembledType(NamedTuple):
    type_: type | ForwardRef
    origin: type | None
    args: Sequence[type]
    type_vars: Sequence["TypeNode"]
    typenode: "TypeNode"


@dataclass
class TypeNode:
    type_: Any
    args: list["TypeNode"] = field(default_factory=list)

    @override
    def __hash__(self) -> int:
        return id(self)


class MISSING:
    pass


class InitOptions(TypedDict):
    slots: bool
    frozen: bool
    init: bool


class UNINITIALIZED:
    pass


class Descriptor(Protocol):
    private_name: str

    def __get__(self, instance: Any | None, owner: type[Any]) -> Self | Any: ...



from typing import NamedTuple

import pytest
from pydantic import ValidationError

from escudeiro.contrib.pydantic import Model, MutableModel
from escudeiro.lazyfields import lazyfield


def test_model_aliases_are_automatically_created_as_camel():
    class Person(Model):
        my_personal_name: str
        type_: str
        id_: int

    assert Person.model_fields["my_personal_name"].alias == "myPersonalName"
    assert Person.model_fields["type_"].alias == "type"
    assert Person.model_fields["id_"].alias == "id"


def test_model_is_frozen():
    class Person(Model):
        name: str

    person = Person(name="Hello")

    with pytest.raises(ValidationError):
        person.name = "world"


def test_model_supports_lazyfields():
    call_count = 0

    class Person(Model):
        name: str
        surname: str

        @lazyfield
        def full_name(self):
            nonlocal call_count
            call_count += 1
            return f"{self.name} {self.surname}"

    person = Person(name="John", surname="Doe")
    assert person.full_name == "John Doe"

    del person.full_name
    person.full_name = "Jane Doe"

    assert person.full_name == "Jane Doe"
    assert call_count == 1


def test_model_populates_by_alias_and_name():
    class Person(Model):
        id_: int
        type_: str

    person = Person(id=1, type_="2")  # pyright: ignore[reportCallIssue]

    assert person.id_ == 1
    assert person.type_ == "2"


def test_model_supports_from_attributes():
    class PersonTuple(NamedTuple):
        name: str
        email: str

    person_tuple = PersonTuple("John Doe", "test@example.com")

    class Person(Model):
        name: str
        email: str

    person = Person.model_validate(person_tuple)

    assert person.name == person_tuple.name
    assert person.email == person_tuple.email


def test_mutable_model_is_not_frozen():
    class Person(MutableModel):
        name: str

    person = Person(name="John Doe")

    person.name = "Jane Doe"

    assert person.name == "Jane Doe"



from __future__ import annotations

import os
from contextlib import contextmanager
from functools import partial
from typing import Any, Self

from escudeiro.data import data
from escudeiro.ds.filetree.helpers import resolve_error
from escudeiro.escudeiro_pyrs import filetree
from escudeiro.exc import FailedFileOperation, InvalidPath
from escudeiro.lazyfields import lazyfield


@data(frozen=False)
class VirtualFileTree:
    root: filetree.FsNode

    @lazyfield
    def _internal(self) -> filetree.FsTree:
        return filetree.FsTree.from_node(self.root)

    @property
    def name(self) -> str:
        return self.root.name

    @classmethod
    def from_basename(cls, name: str) -> Self:
        return cls(filetree.FsNode(name))

    @contextmanager
    def virtual_context(self, dirname: str, *path: str):
        try:
            folder = self._get_node(dirname, *path) or filetree.FsNode(dirname)
            vt = VirtualFileTree(folder)
            yield vt
        except AssertionError as err:
            if "PYTEST_CURRENT_TEST" in os.environ:
                raise
            else:
                raise FailedFileOperation(
                    "unable to complete virtual context after exception", err
                ) from err
        except Exception as err:
            raise FailedFileOperation(
                "unable to complete virtual context after exception", err
            ) from err
        else:
            self.merge(vt, *path)

    def _get_node(self, name: str, *path: str) -> filetree.FsNode | None:
        try:
            parent = self._internal.get_node(*path)
        except ValueError:
            return None
        else:
            return parent.get_shallow(name)

    def merge(self, vt: VirtualFileTree, *path: str) -> None:
        if path:
            *restpath, parent = path
            parent_node = self._internal.create_dir(parent, *restpath)
        else:
            parent_node = self._internal.root
        if curnode := parent_node.get_shallow(vt.name):
            if curnode.is_file():
                raise InvalidPath(
                    f"Trying to create folder {curnode.name} where there is a file."
                )
            else:
                for child in vt.root.children:
                    if curnode.get_shallow(child.name):
                        raise InvalidPath(
                            f"Conflict: {child.name} already exists in {curnode.name}"
                        )
                    curnode.add_child(child)
        else:
            parent_node.add_child(vt.root)

    def create_file(
        self,
        filename: str,
        *path: str,
        content: bytes = b"",
        append: bool = False,
    ) -> filetree.FsNode:
        file_maker = partial(
            self._internal.create_file,
            filename,
            content,
            *path,
        )
        try:
            parent = self._internal.get_node(*path)
        except ValueError as err:
            raise resolve_error(err)
        else:
            node = parent.get_shallow(filename)
            if not node:
                return file_maker()
            elif node.content is None:
                raise InvalidPath(
                    f"Trying to create file {node.name} where there is a folder."
                )
            else:
                writer = node.append_content if append else node.write_content
                writer(content)
                return node

    def create_text_file(
        self,
        filename: str,
        *path: str,
        content: str = "",
        encoding: str = "utf-8",
        append: bool = False,
    ) -> filetree.FsNode:
        return self.create_file(
            filename,
            *path,
            content=content.encode(encoding),
            append=append,
        )

    def create_py_file(
        self,
        filename: str,
        *path: str,
        private: bool = False,
        dunder: bool = False,
        content: str = "",
        append: bool = False,
    ) -> filetree.FsNode:
        return self.create_text_file(
            filetree.python_filename(
                filename,
                dunder=dunder,
                private=private,
            ),
            *path,
            content=content,
            append=append,
        )

    def create_init_file(
        self,
        *path: str,
        content: str = "",
        append: bool = False,
    ) -> filetree.FsNode:
        return self.create_text_file(
            filetree.init_file(),
            *path,
            content=content,
            append=append,
        )

    def create_dir(self, dirname: str, *path: str) -> filetree.FsNode:
        try:
            return self._internal.create_dir(dirname, *path)
        except ValueError as err:
            raise resolve_error(err)

    def get_path(self, pathname: str, *path: str) -> filetree.FsNode:
        try:
            parent = self._internal.get_node(*path)
        except ValueError as e:
            raise resolve_error(e)
        else:
            file = parent.get_shallow(pathname)
            if file is None:
                raise InvalidPath("path not found")
            else:
                return file

    def __parse_dict__(self, by_alias: bool) -> dict[str, Any]:
        output = {}
        stack = [(self.root, output)]

        while stack:
            node, mapping = stack.pop()

            if node.is_file():
                mapping[node.name] = node.content
            else:
                new_mapping = {}
                mapping[node.name] = new_mapping
                for child in node.children:
                    stack.append((child, new_mapping))

        return output
